#
# Bindings.py -- Bindings classes for Ginga viewer.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import os.path
import itertools

from ginga.misc import Bunch, Settings, Callback
from ginga.util.paths import icondir
from ginga import events


class ImageViewBindings(object):
    """Configurable event handling (bindings) for UI events.

    This module handles the mapping of user interface events (keyboard,
    mouse, trackpad, gestures) to operations on a Ginga viewer.
    """

    action_prefixes = ['kp_', 'ms_', 'sc_', 'pi_', 'pa_']

    def __init__(self, logger, settings=None):
        super(ImageViewBindings, self).__init__()

        self.logger = logger

        if settings is None:
            # No settings passed.  Set up defaults.
            settings = Settings.SettingGroup(name='bindings',
                                             logger=self.logger)
            #self.initialize_settings(settings)
        self.settings = settings

        self.features = dict(pan=False, zoom=False, cut=False, cmap=False,
                             flip=False, rotate=False)
        self.cursor_map = {}
        self._modes = {}
        self._cur_mode = None

    def initialize_settings(self, settings):
        settings.add_settings(
            # You should rarely have to change these.
            btn_nobtn=0x0,
            btn_left=0x1,
            btn_middle=0x2,
            btn_right=0x4,
            btn_back=0x8,
            btn_forward=0x10,

            # Set up our standard modifiers
            mod_shift=['shift_l', 'shift_r'],
            mod_ctrl=['control_l', 'control_r'],
            mod_win=['meta_right'],

            default_mode_type='locked',
            default_lock_mode_type='locked',

            # MOUSE/BUTTON
            ms_none=['nobtn'],
            ms_cursor=['left'],
            ms_wheel=[],
        )

    def get_settings(self):
        return self.settings

    def add_mode_obj(self, mode_obj):
        #self.settings.add_defaults(**mode_obj.actions)
        self._modes[str(mode_obj)] = mode_obj

    def get_mode_obj(self, mode_name):
        return self._modes[mode_name]

    def window_map(self, viewer):
        #self.to_default_mode(viewer)
        pass

    def set_bindings(self, viewer):
        viewer.add_callback('map', self.window_map)

        bindmap = viewer.get_bindmap()
        bindmap.clear_button_map()
        bindmap.clear_event_map()

        bindmap.add_callback('mode-set', self.mode_set_cb, viewer)

        # Set up bindings
        self.setup_settings_events(viewer, bindmap)

    def set_mode(self, viewer, name, mode_type='oneshot'):
        bindmap = viewer.get_bindmap()
        bindmap.set_mode(name, mode_type=mode_type)

    def mode_set_cb(self, bm, mode, mode_type, viewer):
        self.logger.debug(f'mode change to {mode}')

        if mode != self._cur_mode:
            if self._cur_mode not in ('meta', None):
                try:
                    self._modes[self._cur_mode].stop()
                except Exception as e:
                    self.logger.error("Error stopping mode '{}': {}".format(self._cur_mode, e),
                                      exc_info=True)

            self._cur_mode = mode
            if self._cur_mode not in ('meta', None):
                try:
                    self._modes[self._cur_mode].start()
                except Exception as e:
                    self.logger.error("Error starting mode '{}': {}".format(self._cur_mode, e),
                                      exc_info=True)

    def parse_combo(self, combo, modes_set, modifiers_set, pfx):
        """
        Parse a string into a mode, a set of modifiers and a trigger.
        """
        mode, mods, trigger = None, set([]), combo
        if '+' in combo:
            if combo.endswith('+'):
                # special case: probably contains the keystroke '+'
                trigger, combo = '+', combo[:-1]
                if '+' in combo:
                    items = set(combo.split('+'))
                else:
                    items = set(combo)
            else:
                # trigger is always specified last
                items = combo.split('+')
                trigger, items = items[-1], set(items[:-1])

            if '*' in items:
                items.remove('*')
                # modifier wildcard
                mods = '*'
            else:
                mods = items.intersection(modifiers_set)

            mode = items.intersection(modes_set)
            if len(mode) == 0:
                mode = None
            else:
                mode = mode.pop()

        if pfx is not None:
            trigger = pfx + trigger

        return (mode, mods, trigger)

    def setup_settings_events(self, viewer, bindmap):

        self.initialize_settings(self.settings)

        # merge in all default settings from various modes, being
        # careful not to override any user preferences
        act_d = dict()
        from ginga.modes.modeinfo import available_modes
        for mode_class in available_modes:
            mode_obj = mode_class(viewer, settings=self.settings)
            for name, value in mode_obj.actions.items():
                pfx = name[:3]
                if pfx not in self.action_prefixes:
                    act_d.setdefault(name, value)
                    continue
                # a binding
                act_l = act_d.setdefault(name, [])
                for trigger in value:
                    if trigger not in act_l:
                        act_l.append(trigger)
            self.add_mode_obj(mode_obj)

        self.settings.add_defaults(**act_d)

        # Now settings should have all available modes defined
        d = self.settings.get_dict()

        # First scan settings for buttons and modes
        bindmap.clear_button_map()
        bindmap.clear_modifier_map()
        bindmap.clear_mode_map()
        bindmap.clear_event_map()

        mode_type = self.settings.get('default_mode_type', 'oneshot')
        bindmap.set_default_mode_type(mode_type)

        for name, value in d.items():
            if name.startswith('mod_'):
                modname = name[4:]
                for combo in value:
                    # NOTE: for now no chorded combinations to make modifiers
                    keyname = combo
                    bindmap.add_modifier(keyname, modname)

            elif name.startswith('cur_'):
                curname = name[4:]
                self.add_cursor(viewer, curname, value)

            elif name.startswith('btn_'):
                btnname = name[4:]
                bindmap.map_button(value, btnname)

            elif name.startswith('dmod_'):
                mode_name = name[5:]
                keyname, mode_type, curname = value
                bindmap.add_mode(keyname, mode_name, mode_type=mode_type,
                                 msg=None)
                if curname is not None:
                    self.cursor_map[mode_name] = curname

        self.merge_actions(viewer, bindmap, self, d.items())

        # merge in specific actions for each mode
        for mode_obj in self._modes.values():
            actions = mode_obj.actions
            d = {key: self.settings[key]
                 for key in actions.keys() if key in self.settings}
            self.merge_actions(viewer, bindmap, mode_obj, d.items())

    def merge_actions(self, viewer, bindmap, obj, tups):

        modes_set = bindmap.get_modes()
        modifiers_set = bindmap.get_modifiers()

        # Add events
        for name, value in tups:
            if len(name) <= 3:
                continue

            pfx = name[:3]
            if pfx not in self.action_prefixes:
                continue

            evname = name[3:]
            for combo in value:
                mode, modifiers, trigger = self.parse_combo(combo, modes_set,
                                                            modifiers_set, pfx)
                if modifiers == '*':
                    # wildcard; register for all modifier combinations
                    modifiers_poss = set([])
                    for i in range(len(modifiers_set) + 1):
                        modifiers_poss = modifiers_poss.union(
                            itertools.combinations(modifiers_set, i))
                    for modifiers in modifiers_poss:
                        bindmap.map_event(mode, modifiers, trigger, evname)
                else:
                    bindmap.map_event(mode, modifiers, trigger, evname)

            # Register for this symbolic event if we have a handler for it
            try:
                cb_method = getattr(obj, name)

            except AttributeError:
                # Do we need a warning here?
                #self.logger.warning("No method found matching '%s'" % (name))
                cb_method = None

            if pfx == 'kp_':
                # keyboard event
                event = 'keydown-%s' % (evname)
                viewer.enable_callback(event)
                if cb_method is not None and viewer.num_callbacks(event) == 0:
                    viewer.add_callback(event, cb_method)

            elif pfx == 'ms_':
                # mouse/button event
                for action in ('down', 'move', 'up'):
                    event = '%s-%s' % (evname, action)
                    viewer.enable_callback(event)
                    if cb_method is not None and viewer.num_callbacks(event) == 0:
                        viewer.add_callback(event, cb_method)

            elif pfx == 'sc_':
                # scrolling event
                event = '%s-scroll' % evname
                viewer.enable_callback(event)
                if cb_method is not None and viewer.num_callbacks(event) == 0:
                    viewer.add_callback(event, cb_method)

            elif pfx == 'pi_':
                # pinch event
                event = '%s-pinch' % evname
                viewer.enable_callback(event)
                if cb_method is not None and viewer.num_callbacks(event) == 0:
                    viewer.add_callback(event, cb_method)

            elif pfx == 'pa_':
                # pan event
                event = '%s-pan' % evname
                viewer.enable_callback(event)
                if cb_method is not None and viewer.num_callbacks(event) == 0:
                    viewer.add_callback(event, cb_method)

    def reset(self, viewer):
        bindmap = viewer.get_bindmap()
        bindmap.reset_mode(viewer)
        viewer.onscreen_message(None)

    def add_cursor(self, viewer, curname, curpath):
        if not curpath.startswith('/'):
            curpath = os.path.join(icondir, curpath)
        cursor = viewer.make_cursor(curpath, 8, 8, size=(16, 16))
        viewer.define_cursor(curname, cursor)

    #####  ENABLERS #####
    # These methods are a quick way to enable or disable certain user
    # interface features in a ImageView window

    def enable_pan(self, tf):
        """Enable the image to be panned interactively (True/False)."""
        self.enable(pan=tf)

    def enable_zoom(self, tf):
        """Enable the image to be zoomed interactively (True/False)."""
        self.enable(zoom=tf)

    def enable_cuts(self, tf):
        """Enable the cuts levels to be set interactively (True/False)."""
        self.enable(cut=tf)

    def enable_cmap(self, tf):
        """Enable the color map to be warped interactively (True/False)."""
        self.enable(cmap=tf)

    def enable_flip(self, tf):
        """Enable the image to be flipped interactively (True/False)."""
        self.enable(flip=tf)

    def enable_rotate(self, tf):
        """Enable the image to be rotated interactively (True/False)."""
        self.enable(rotate=tf)

    def enable(self, **kwargs):
        """General enable function encompassing all user interface features.
        Usage (e.g.): viewer.enable(rotate=False, flip=True)
        """
        for feat, value in kwargs.items():
            feat = feat.lower()
            if feat not in self.features:
                raise ValueError("'%s' is not a feature. Must be one of %s" % (
                    feat, str(list(self.features.keys()))))

            self.features[feat] = bool(value)

    def enable_all(self, tf):
        for feat in list(self.features.keys()):
            self.features[feat] = tf

    def get_feature_allow(self, feat_name):
        return self.features[feat_name]

    def get_direction(self, direction, rev=False):
        """Translate a direction in compass degrees into 'up' or 'down'."""
        # TODO: merge with the one in mode_base
        if (direction < 90.0) or (direction >= 270.0):
            if not rev:
                return 'up'
            else:
                return 'down'
        elif (90.0 <= direction < 270.0):
            if not rev:
                return 'down'
            else:
                return 'up'
        else:
            return 'none'


class BindingMapError(Exception):
    pass


class BindingMapper(Callback.Callbacks):
    """Map physical events to logical events.

    The BindingMapper class maps physical events (key presses, button
    clicks, mouse movement, etc) into logical events.  By registering for
    logical events, plugins and other event handling code doesn't need to
    care about the physical controls bindings.  The bindings can be changed
    and everything continues to work.
    """

    def __init__(self, logger, btnmap=None, mode_map=None, modifier_map=None):
        Callback.Callbacks.__init__(self)

        self.logger = logger

        # For event mapping
        self.event_names = ['keydown', 'keyup',
                            'btn-down', 'btn-move', 'btn-up',
                            'scroll', 'pinch', 'pan']
        self.eventmap = {}

        self._kbdmode = None
        self._kbdmode_types = ('held', 'oneshot', 'locked')
        self._kbdmode_type = 'held'
        self._kbdmode_type_default = 'locked'
        self._delayed_reset = False
        self._modifiers = frozenset([])

        # Set up button mapping
        if btnmap is None:
            btnmap = {0x1: 'cursor', 0x2: 'wheel', 0x4: 'draw'}
        self.btnmap = btnmap
        self._button = 0

        # Set up modifier mapping
        if modifier_map is None:
            self.modifier_map = {}
            for keyname in ('shift_l', 'shift_r'):
                self.add_modifier(keyname, 'shift')
            for keyname in ('control_l', 'control_r'):
                self.add_modifier(keyname, 'ctrl')
            for keyname in ('meta_right',):
                self.add_modifier(keyname, 'win')
        else:
            self.modifier_map = modifier_map

        # Set up mode mapping
        if mode_map is None:
            self.mode_map = {}
        else:
            self.mode_map = mode_map

        self._empty_set = frozenset([])
        self.mode_tbl = dict()

        # For callbacks
        for name in ('mode-set', ):
            self.enable_callback(name)

    def add_modifier(self, keyname, modname):
        bnch = Bunch.Bunch(name=modname)
        self.modifier_map[keyname] = bnch
        self.modifier_map['mod_%s' % modname] = bnch

    def get_modifiers(self):
        return set([bnch.name for keyname, bnch
                    in self.modifier_map.items()])

    def clear_modifier_map(self):
        self.modifier_map = {}

    def set_mode_map(self, mode_map):
        self.mode_map = mode_map

    def clear_mode_map(self):
        self.mode_map = {}

    def has_mode(self, mode_name):
        key = 'mode_%s' % mode_name
        return key in self.mode_map

    def remove_mode(self, mode_name):
        key = 'mode_%s' % mode_name
        del self.mode_map[key]

    def current_mode(self):
        return (self._kbdmode, self._kbdmode_type)

    def get_modes(self):
        return set([bnch.name for keyname, bnch in self.mode_map.items()])

    def add_mode(self, keyname, mode_name, mode_type='held', msg=None):
        if mode_type is not None:
            if mode_type not in self._kbdmode_types:
                raise ValueError("Bad mode type '%s': must be one of %s" % (
                    mode_type, self._kbdmode_types))

        bnch = Bunch.Bunch(name=mode_name, type=mode_type, msg=msg)
        if keyname is not None:
            # Key to launch this mode
            if keyname[0:2] == '__':
                keyname = keyname[2]
                self.mode_tbl[keyname] = bnch
            else:
                self.mode_map[keyname] = bnch
        self.mode_map['mode_%s' % mode_name] = bnch

    def set_mode(self, name, mode_type=None):
        if mode_type is None:
            mode_type = self._kbdmode_type_default
        if mode_type not in self._kbdmode_types:
            raise ValueError("Bad mode type '%s': must be one of %s" % (
                mode_type, self._kbdmode_types))
        self._kbdmode = name
        if name is None:
            # like a reset_mode()
            mode_type = 'held'
            self._delayed_reset = False
        self._kbdmode_type = mode_type
        self.logger.info("set keyboard mode to '%s' type=%s" % (name, mode_type))
        self.make_callback('mode-set', self._kbdmode, self._kbdmode_type)

    def set_default_mode_type(self, mode_type):
        if mode_type not in self._kbdmode_types:
            raise ValueError("Bad mode type '%s': must be one of %s" % (
                mode_type, self._kbdmode_types))
        self._kbdmode_type_default = mode_type

    def get_default_mode_type(self):
        return self._kbdmode_type_default

    def reset_mode(self, viewer):
        try:
            bnch = self.mode_map['mode_%s' % self._kbdmode]
        except Exception:
            bnch = None
        self._kbdmode = None
        self._kbdmode_type = 'held'
        self._delayed_reset = False
        self.logger.debug("set keyboard mode reset")
        # clear onscreen message, if any
        if (bnch is not None) and (bnch.msg is not None):
            viewer.onscreen_message(None)
        self.make_callback('mode-set', self._kbdmode, self._kbdmode_type)

    def clear_button_map(self):
        self.btnmap = {}

    def map_button(self, btncode, alias):
        """For remapping the buttons to different names. 'btncode' is a
        fixed button code and 'alias' is a logical name.
        """
        self.btnmap[btncode] = alias

    def get_buttons(self):
        return set([alias for keyname, alias in self.btnmap.items()])

    def get_button(self, btncode):
        return self.btnmap.get(btncode, None)

    def clear_event_map(self):
        self.eventmap = {}

    def map_event(self, mode, modifiers, trigger, eventname):
        self.eventmap[(mode, frozenset(tuple(modifiers)),
                       trigger)] = Bunch.Bunch(name=eventname)

    def register_for_events(self, viewer):
        # Add callbacks for interesting events
        viewer.add_callback('motion', self.window_motion)
        viewer.add_callback('button-press', self.window_button_press)
        viewer.add_callback('button-release', self.window_button_release)
        viewer.add_callback('key-press', self.window_key_press)
        viewer.add_callback('key-release', self.window_key_release)
        ## viewer.add_callback('drag-drop', self.window_drag_drop)
        viewer.add_callback('scroll', self.window_scroll)
        viewer.add_callback('map', self.window_map)
        viewer.add_callback('focus', self.window_focus)
        viewer.add_callback('enter', self.window_enter)
        viewer.add_callback('leave', self.window_leave)
        if viewer.has_callback('pinch'):
            viewer.add_callback('pinch', self.window_pinch)
        if viewer.has_callback('pan'):
            viewer.add_callback('pan', self.window_pan)

        for pfx in self.event_names:
            # TODO: add moded versions of callbacks?
            viewer.enable_callback('%s-none' % (pfx))

    def mode_key_down(self, viewer, keyname):
        """This method is called when a key is pressed and was not handled
        by some other handler with precedence, such as a subcanvas.
        """
        # Is this a mode key?
        if keyname not in self.mode_map:
            if (self._kbdmode != 'meta') or (keyname not in self.mode_tbl):
                # No
                return False
            bnch = self.mode_tbl[keyname]
        else:
            bnch = self.mode_map[keyname]

        mode_name = bnch.name
        self.logger.debug("cur mode='%s' mode pressed='%s'" % (
            self._kbdmode, mode_name))

        if self._delayed_reset:
            # <== this shouldn't happen, but put here to reset handling
            # of delayed_reset just in case (see cursor up handling)
            self._delayed_reset = False
            return True

        if ((self._kbdmode in (None, 'meta')) or
            (self._kbdmode_type != 'locked') or (mode_name == 'meta')):
            if self._kbdmode is not None:
                self.reset_mode(viewer)

            # activate this mode
            if self._kbdmode in (None, 'meta'):
                mode_type = bnch.type
                if mode_type is None:
                    mode_type = self._kbdmode_type_default
                self.set_mode(mode_name, mode_type)
                if bnch.msg is not None:
                    viewer.onscreen_message(bnch.msg)

                return True

        return False

    def mode_key_up(self, viewer, keyname):
        """This method is called when a key is pressed in a mode and was
        not handled by some other handler with precedence, such as a
        subcanvas.
        """
        # Is this a mode key?
        if keyname not in self.mode_map:
            # <== no
            return False

        bnch = self.mode_map[keyname]
        if self._kbdmode == bnch.name:
            # <-- the current mode key is being released
            if bnch.type == 'held':
                if self._button == 0:
                    # if no button is being held, then reset mode
                    self.reset_mode(viewer)
                else:
                    self._delayed_reset = True
            return True

        return False

    def window_map(self, viewer):
        return True

    def window_focus(self, viewer, has_focus):
        if not has_focus:
            # fixes a problem with not receiving key release events when the
            # window loses focus
            self._modifiers = frozenset([])
        return False

    def window_enter(self, viewer):
        return False

    def window_leave(self, viewer):
        return False

    def window_key_press(self, viewer, keyname):
        self.logger.debug("keyname=%s" % (keyname))
        # Is this a modifer key?
        if keyname in self.modifier_map:
            bnch = self.modifier_map[keyname]
            self._modifiers = self._modifiers.union(set([bnch.name]))
            return True

        if self.mode_key_down(viewer, keyname):
            return True

        trigger = 'kp_' + keyname
        last_x, last_y = viewer.get_last_data_xy()

        event = events.KeyEvent(key=keyname, state='down', mode=self._kbdmode,
                                modifiers=self._modifiers, viewer=viewer,
                                data_x=last_x, data_y=last_y)

        if self._kbdmode is None:
            cbname = 'key-down-none'

        else:
            if keyname == 'escape':
                idx = (None, self._empty_set, trigger)
            else:
                idx = (self._kbdmode, self._modifiers, trigger)

            if idx in self.eventmap:
                emap = self.eventmap[idx]
                cbname = 'keydown-%s' % (emap.name)
            else:
                # TEMP: hack to get around the issue of how keynames
                # are generated--shifted characters with no modifiers
                idx = (self._kbdmode, self._empty_set, trigger)
                if idx in self.eventmap:
                    emap = self.eventmap[idx]
                    cbname = 'keydown-%s' % (emap.name)
                else:
                    cbname = 'key-down-%s' % str(self._kbdmode).lower()

        res = viewer.make_ui_callback_viewer(viewer, cbname, event,
                                             last_x, last_y)

        if not event.was_handled() and not res:
            # no response for this canvas or mode, try non-mode entry
            idx = (None, self._modifiers, trigger)
            if idx in self.eventmap:
                emap = self.eventmap[idx]
                cbname = 'keydown-%s' % (emap.name)
                res = viewer.make_ui_callback_viewer(viewer, cbname, event,
                                                     last_x, last_y)
            else:
                # TEMP: hack to get around the issue of how keynames
                # are generated--shifted characters with no modifiers
                idx = (None, self._empty_set, trigger)
                if idx in self.eventmap:
                    emap = self.eventmap[idx]
                    cbname = 'keydown-%s' % (emap.name)
                    res = viewer.make_ui_callback_viewer(viewer, cbname, event,
                                                         last_x, last_y)

        return res

    def window_key_release(self, viewer, keyname):
        self.logger.debug("keyname=%s" % (keyname))

        # Is this a modifer key?
        if keyname in self.modifier_map:
            bnch = self.modifier_map[keyname]
            self._modifiers = self._modifiers.difference(set([bnch.name]))
            return True

        if self.mode_key_up(viewer, keyname):
            return True

        trigger = 'kp_' + keyname
        last_x, last_y = viewer.get_last_data_xy()

        event = events.KeyEvent(key=keyname, state='up', mode=self._kbdmode,
                                modifiers=self._modifiers, viewer=viewer,
                                data_x=last_x, data_y=last_y)

        if self._kbdmode is None:
            cbname = 'key-up-none'

        else:
            idx = (self._kbdmode, self._modifiers, trigger)
            if idx in self.eventmap:
                emap = self.eventmap[idx]
                cbname = 'keyup-%s' % (emap.name)

            else:
                # TEMP: hack to get around the issue of how keynames
                # are generated--shifted characters with no modifiers
                idx = (self._kbdmode, self._empty_set, trigger)
                if idx in self.eventmap:
                    emap = self.eventmap[idx]
                    cbname = 'keyup-%s' % (emap.name)

                else:
                    cbname = 'key-up-%s' % str(self._kbdmode).lower()

        res = viewer.make_ui_callback_viewer(viewer, cbname, event,
                                             last_x, last_y)

        if not event.was_handled() and not res:
            # no response for this canvas or mode, try non-mode entry
            idx = (None, self._modifiers, trigger)
            if idx in self.eventmap:
                emap = self.eventmap[idx]
                cbname = 'keyup-%s' % (emap.name)
                res = viewer.make_ui_callback_viewer(viewer, cbname, event,
                                                     last_x, last_y)
            else:
                # TEMP: hack to get around the issue of how keynames
                # are generated--shifted characters with no modifiers
                idx = (None, self._empty_set, trigger)
                if idx in self.eventmap:
                    emap = self.eventmap[idx]
                    cbname = 'keyup-%s' % (emap.name)
                    res = viewer.make_ui_callback_viewer(viewer, cbname, event,
                                                         last_x, last_y)
        return res

    def window_button_press(self, viewer, btncode, data_x, data_y):
        self.logger.debug("x,y=%d,%d btncode=%s" % (data_x, data_y,
                                                    hex(btncode)))
        self._button |= btncode
        button = self.get_button(btncode)
        if button is None:
            self.logger.error("unrecognized button code (%x)" % (btncode))
            return False

        trigger = 'ms_' + button
        event = events.PointEvent(button=button, state='down',
                                  mode=self._kbdmode,
                                  modifiers=self._modifiers, viewer=viewer,
                                  data_x=data_x, data_y=data_y)

        if self._kbdmode is None:
            cbname = 'btn-down-none'

        else:
            idx = (self._kbdmode, self._modifiers, trigger)
            if idx in self.eventmap:
                emap = self.eventmap[idx]
                cbname = '%s-down' % (emap.name)

            else:
                cbname = 'btn-down-%s' % str(self._kbdmode).lower()

        self.logger.debug("making callback for %s (mode=%s)" % (
            cbname, self._kbdmode))

        res = viewer.make_ui_callback_viewer(viewer, cbname, event,
                                             data_x, data_y)

        if not event.was_handled() and not res:
            # no entry for this mode, try non-mode entry
            idx = (None, self._modifiers, trigger)
            if idx in self.eventmap:
                emap = self.eventmap[idx]
                cbname = '%s-down' % (emap.name)
                res = viewer.make_ui_callback_viewer(viewer, cbname, event,
                                                     data_x, data_y)
        return res

    def window_motion(self, viewer, btncode, data_x, data_y):

        button = self.get_button(btncode)
        if button is None:
            self.logger.error("unrecognized button code (%x)" % (btncode))
            return False

        trigger = 'ms_' + button
        event = events.PointEvent(button=button, state='move',
                                  mode=self._kbdmode,
                                  modifiers=self._modifiers, viewer=viewer,
                                  data_x=data_x, data_y=data_y)

        if self._kbdmode is None:
            cbname = 'btn-move-none'

        else:
            idx = (self._kbdmode, self._modifiers, trigger)
            if idx in self.eventmap:
                emap = self.eventmap[idx]
                cbname = '%s-move' % (emap.name)

            else:
                cbname = 'btn-move-%s' % str(self._kbdmode).lower()

        res = viewer.make_ui_callback_viewer(viewer, cbname, event,
                                             data_x, data_y)

        if not event.was_handled() and not res:
            # no entry for this mode, try non-mode entry
            idx = (None, self._modifiers, trigger)
            if idx in self.eventmap:
                emap = self.eventmap[idx]
                cbname = '%s-move' % (emap.name)
                res = viewer.make_ui_callback_viewer(viewer, cbname, event,
                                                     data_x, data_y)
        return res

    def window_button_release(self, viewer, btncode, data_x, data_y):
        self.logger.debug("x,y=%d,%d button=%s" % (data_x, data_y,
                                                   hex(btncode)))
        self._button &= ~btncode
        button = self.get_button(btncode)
        if button is None:
            self.logger.error("unrecognized button code (%x)" % (btncode))
            return False

        trigger = 'ms_' + button
        event = events.PointEvent(button=button, state='up', mode=self._kbdmode,
                                  modifiers=self._modifiers, viewer=viewer,
                                  data_x=data_x, data_y=data_y)

        if self._kbdmode is None:
            cbname = 'btn-up-none'

        else:
            idx = (self._kbdmode, self._modifiers, trigger)
            if idx in self.eventmap:
                # release mode if this is a oneshot mode
                if (self._kbdmode_type == 'oneshot') or (self._delayed_reset):
                    self.reset_mode(viewer)
                emap = self.eventmap[idx]
                cbname = '%s-up' % (emap.name)

            else:
                cbname = 'btn-up-%s' % str(self._kbdmode).lower()

        res = viewer.make_ui_callback_viewer(viewer, cbname, event,
                                             data_x, data_y)

        if not event.was_handled() and not res:
            # no entry for this mode, try non-mode entry
            idx = (None, self._modifiers, trigger)
            if idx in self.eventmap:
                emap = self.eventmap[idx]
                cbname = '%s-up' % (emap.name)
                res = viewer.make_ui_callback_viewer(viewer, cbname, event,
                                                     data_x, data_y)
        return res

    def window_scroll(self, viewer, direction, amount, data_x, data_y):
        trigger = 'sc_scroll'
        event = events.ScrollEvent(button='scroll', state='scroll',
                                   mode=self._kbdmode,
                                   modifiers=self._modifiers, viewer=viewer,
                                   direction=direction, amount=amount,
                                   data_x=data_x, data_y=data_y)

        if self._kbdmode is None:
            cbname = 'scroll-none'

        else:
            idx = (self._kbdmode, self._modifiers, trigger)
            if idx in self.eventmap:
                emap = self.eventmap[idx]
                cbname = '%s-scroll' % (emap.name)

            else:
                cbname = 'scroll-%s' % str(self._kbdmode).lower()

        res = viewer.make_ui_callback_viewer(viewer, cbname, event)

        if not event.was_handled() and not res:
            # no entry for this mode, try non-mode entry
            idx = (None, self._modifiers, trigger)
            if idx in self.eventmap:
                emap = self.eventmap[idx]
                cbname = '%s-scroll' % (emap.name)
                res = viewer.make_ui_callback_viewer(viewer, cbname, event)

        return res

    def window_pinch(self, viewer, state, rot_deg, scale):
        btncode = 0
        button = self.get_button(btncode)
        if button is None:
            self.logger.error("unrecognized button code (%x)" % (btncode))
            return False

        trigger = 'pi_pinch'
        last_x, last_y = viewer.get_last_data_xy()
        event = events.PinchEvent(button=button, state=state,
                                  mode=self._kbdmode,
                                  modifiers=self._modifiers, viewer=viewer,
                                  rot_deg=rot_deg, scale=scale,
                                  data_x=last_x, data_y=last_y)

        if self._kbdmode is None:
            cbname = 'pinch-none'

        else:
            idx = (self._kbdmode, self._modifiers, trigger)
            if idx in self.eventmap:
                emap = self.eventmap[idx]
                cbname = '%s-pinch' % (emap.name)

            else:
                cbname = 'pinch-%s' % (str(self._kbdmode).lower())

        self.logger.debug("making callback for %s (mode=%s)" % (
            cbname, self._kbdmode))
        res = viewer.make_ui_callback_viewer(viewer, cbname, event)

        if not event.was_handled() and not res:
            # no entry for this mode, try non-mode entry
            idx = (None, self._modifiers, trigger)
            if idx in self.eventmap:
                emap = self.eventmap[idx]
                cbname = '%s-pinch' % (emap.name)
                res = viewer.make_ui_callback_viewer(viewer, cbname, event)

        return res

    def window_pan(self, viewer, state, delta_x, delta_y):
        btncode = 0
        button = self.get_button(btncode)
        if button is None:
            self.logger.error("unrecognized button code (%x)" % (btncode))
            return False

        trigger = 'pa_pan'
        last_x, last_y = viewer.get_last_data_xy()
        event = events.PanEvent(button=button, state=state, mode=self._kbdmode,
                                modifiers=self._modifiers, viewer=viewer,
                                delta_x=delta_x, delta_y=delta_y,
                                data_x=last_x, data_y=last_y)

        if self._kbdmode is None:
            cbname = 'pan-none'

        else:
            idx = (self._kbdmode, self._modifiers, trigger)
            if idx in self.eventmap:
                emap = self.eventmap[idx]
                cbname = '%s-pan' % (emap.name)

            else:
                cbname = 'pan-%s' % (str(self._kbdmode).lower())

        self.logger.debug("making callback for %s (mode=%s)" % (
            cbname, self._kbdmode))
        res = viewer.make_ui_callback_viewer(viewer, cbname, event)

        if not event.was_handled() and not res:
            # no entry for this mode, try non-mode entry
            idx = (None, self._modifiers, trigger)
            if idx in self.eventmap:
                emap = self.eventmap[idx]
                cbname = '%s-pan' % (emap.name)
                res = viewer.make_ui_callback_viewer(viewer, cbname, event)

        return res



#! /usr/bin/env python
#
# example1_mpl.py -- Simple, configurable FITS viewer using a matplotlib
#                      QtAgg backend for Ginga and embedded in a Qt program.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
"""
Usage:
   example1_mpl.py [fits file]

You need Qt5 or Qt6 with pyqt bindings (or pyside) installed to run this
example.
"""

import sys

from matplotlib.figure import Figure

from ginga.qtw.QtHelp import QtGui, QtCore
from ginga.qtw import QtHelp
from ginga.mplw.ImageViewMpl import CanvasView
from ginga.mplw.FigureCanvasQt import FigureCanvas
from ginga.misc import log
from ginga.util.loader import load_data


class FitsViewer(QtGui.QMainWindow):

    def __init__(self, logger):
        super(FitsViewer, self).__init__()
        self.logger = logger

        fig = Figure()
        w = FigureCanvas(fig)

        fi = CanvasView(logger=self.logger)
        fi.enable_autocuts('on')
        fi.set_autocut_params('zscale')
        fi.enable_auto_orient(True)
        fi.enable_autozoom('on')
        #fi.set_callback('drag-drop', self.drop_file)
        fi.set_callback('none-move', self.motion)
        fi.set_bg(0.2, 0.2, 0.2)
        fi.show_mode_indicator(True, corner='ur')
        fi.ui_set_active(True)
        fi.set_figure(fig)
        fi.show_mode_indicator(True, corner='ur')
        self.fitsimage = fi

        fi.get_bindings().enable_all(True)

        vbox = QtGui.QVBoxLayout()
        vbox.setContentsMargins(QtCore.QMargins(2, 2, 2, 2))
        vbox.setSpacing(1)

        vbox.addWidget(w, stretch=1)

        self.readout = QtGui.QLabel("")
        vbox.addWidget(self.readout, stretch=0,
                       alignment=QtCore.Qt.AlignCenter)

        hbox = QtGui.QHBoxLayout()
        hbox.setContentsMargins(QtCore.QMargins(4, 2, 4, 2))

        wopen = QtGui.QPushButton("Open File")
        wopen.clicked.connect(self.open_file)
        wquit = QtGui.QPushButton("Quit")
        wquit.clicked.connect(self.close)

        hbox.addStretch(1)
        for w in (wopen, wquit):
            hbox.addWidget(w, stretch=0)

        hw = QtGui.QWidget()
        hw.setLayout(hbox)
        vbox.addWidget(hw, stretch=0)

        vw = QtGui.QWidget()
        self.setCentralWidget(vw)
        vw.setLayout(vbox)

        fi.configure(512, 512)

    def get_fitsimage(self):
        return self.fitsimage

    def load_file(self, filepath):
        image = load_data(filepath, logger=self.logger)
        self.fitsimage.set_image(image)

    def open_file(self):
        res = QtGui.QFileDialog.getOpenFileName(self, "Open FITS file",
                                                ".", "FITS files (*.fits)")
        if isinstance(res, tuple):
            fileName = res[0]
        else:
            fileName = str(res)
        if len(fileName) != 0:
            self.load_file(fileName)

    def drop_file(self, fitsimage, paths):
        fileName = paths[0]
        self.load_file(fileName)

    def motion(self, fitsimage, button, data_x, data_y):

        # Get the value under the data coordinates
        try:
            #value = fitsimage.get_data(data_x, data_y)
            # We report the value across the pixel, even though the coords
            # change halfway across the pixel
            value = fitsimage.get_data(int(data_x + 0.5), int(data_y + 0.5))

        except Exception:
            value = None

        fits_x, fits_y = data_x + 1, data_y + 1

        # Calculate WCS RA
        try:
            # NOTE: image function operates on DATA space coords
            image = fitsimage.get_image()
            if image is None:
                # No image loaded
                return
            ra_txt, dec_txt = image.pixtoradec(fits_x, fits_y,
                                               format='str', coords='fits')
        except Exception as e:
            self.logger.warning("Bad coordinate conversion: %s" % (
                str(e)))
            ra_txt = 'BAD WCS'
            dec_txt = 'BAD WCS'

        text = "RA: %s  DEC: %s  X: %.2f  Y: %.2f  Value: %s" % (
            ra_txt, dec_txt, fits_x, fits_y, value)
        self.readout.setText(text)


def main(options, args):

    if QtHelp.have_pyqt6 or QtHelp.have_pyside6:
        QtGui.QApplication.setHighDpiScaleFactorRoundingPolicy(
            QtCore.Qt.HighDpiScaleFactorRoundingPolicy.Floor)
    app = QtGui.QApplication(args)

    logger = log.get_logger(name="example1", options=options)
    viewer = FitsViewer(logger)
    viewer.resize(524, 540)
    viewer.show()
    app.setActiveWindow(viewer)

    if len(args) > 0:
        viewer.load_file(args[0])

    app.exec_()


if __name__ == "__main__":

    # Parse command line options
    from argparse import ArgumentParser

    argprs = ArgumentParser()

    argprs.add_argument("--debug", dest="debug", default=False,
                        action="store_true",
                        help="Enter the pdb debugger on main()")
    argprs.add_argument("--profile", dest="profile", action="store_true",
                        default=False,
                        help="Run the profiler on main()")
    log.addlogopts(argprs)

    (options, args) = argprs.parse_known_args(sys.argv[1:])

    # Are we debugging this?
    if options.debug:
        import pdb

        pdb.run('main(options, args)')

    # Are we profiling this?
    elif options.profile:
        import profile

        print(("%s profile:" % sys.argv[0]))
        profile.run('main(options, args)')

    else:
        main(options, args)

# END



#
# json.py -- augmented JSON parsing
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import json

from ginga.misc import Bunch

__all__ = ['BunchEncoder', 'as_bunch', 'dumps', 'loads']


class BunchEncoder(json.JSONEncoder):
    """Custom encoder to serialize Ginga's Bunch.Bunch class.

    Usage
    -----
    st = json.dumps(objs, indent=2, cls=BunchEncoder)

    """
    def default(self, obj):
        if isinstance(obj, Bunch.Bunch):
            d = dict(__bunch__=True)
            d.update(obj)
            return d
        return json.JSONEncoder.default(self, obj)


def as_bunch(dct):
    """If `dct` is a serialized Bunch, return it as a Bunch."""
    if '__bunch__' in dct:
        d = dct.copy()
        del d['__bunch__']
        return Bunch.Bunch(d)
    return dct


def dumps(*args, **kwargs):
    """Like json.dumps(), but also serializes Ginga Bunch.Bunch type."""
    d = dict(cls=BunchEncoder)
    d.update(kwargs)
    return json.dumps(*args, **d)


def loads(*args, **kwargs):
    """Like for json.loads(), but also deserializes Ginga Bunch.Bunch type."""
    d = dict(object_hook=as_bunch)
    d.update(kwargs)
    return json.loads(*args, **d)



#
# render.py -- base class and utilities for Ginga renderers
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.

from io import BytesIO

import numpy as np
from PIL import Image

from ginga import trcalc, colors
from ginga.fonts import font_asst
from ginga.util import pipeline
from ginga.util.stages import render
from ginga.misc import Bunch


class RenderError(Exception):
    """Base class for exceptions thrown by canvas renderers."""
    pass


class Line:
    def __init__(self, color='black', alpha=1.0, linewidth=1,
                 linestyle='solid'):
        if color is None:
            color = 'black'
        self.color = color
        self.alpha = alpha
        self.linewidth = linewidth
        self.linestyle = linestyle

        # resolve the color to a 4-tuple of (r, g, b, a)
        self._color_4tup = get_color(self.color, self.alpha)

        # attributes in here can be set by the renderer
        self.render = Bunch.Bunch()

    def get_bpp_color(self, bpp):
        r, g, b, a = self._color_4tup
        maxv = 2**bpp - 1
        return (int(r * maxv), int(g * maxv), int(b * maxv), int(a * maxv))


class Fill:
    def __init__(self, color=None, alpha=1.0):
        self.color = color
        self.alpha = alpha

        # resolve the color to a 4-tuple of (r, g, b, a)
        self._color_4tup = get_color(self.color, self.alpha)

        # attributes in here can be set by the renderer
        self.render = Bunch.Bunch()

    def get_bpp_color(self, bpp):
        r, g, b, a = self._color_4tup
        maxv = 2**bpp - 1
        return (int(r * maxv), int(g * maxv), int(b * maxv), int(a * maxv))


class Font:
    def __init__(self, fontname='sans', fontsize=12):
        if fontname is None:
            fontname = 'Sans Serif'
        fontname = font_asst.resolve_alias(fontname, fontname)
        self.fontname = fontname
        if fontsize is None:
            fontsize = 12
        self.fontsize = fontsize
        # scale relative to a 12pt font
        self.scale = fontsize / 12.0

        # attributes in here can be set by the renderer
        self.render = Bunch.Bunch()


class RenderContextBase:
    """Base class from which all RenderContext classes are derived."""

    def __init__(self, renderer, viewer):
        self.renderer = renderer
        self.viewer = viewer
        self.renderkey = renderer.kind

        self.line = None
        self.fill = None
        self.font = None

        # special value that can vary by subclasses to scale the font
        # according to each renderer type
        self._font_scale_factor = 1.0

    def __get_color(self, color, alpha):
        return get_color(color, alpha)

    def get_line(self, color, alpha=1.0, linewidth=1, linestyle='solid'):
        return Line(color=color, linewidth=linewidth,
                    linestyle=linestyle, alpha=alpha)

    def get_fill(self, color, alpha=1.0):
        return Fill(color=color, alpha=alpha)

    def get_font(self, fontname, fontsize=None, fontscale=False,
                 fontsize_min=2, fontsize_max=None):
        fontname = font_asst.resolve_alias(fontname, fontname)
        if fontsize is None:
            fontsize = 12
            fontsize = self.scale_fontsize(fontsize)
        if fontscale:
            # requesting to scale the font according to the viewer zoom level
            scale = self.viewer.get_scale_max()
            n = 1.4
            fontsize = max(fontsize_min, fontsize + np.log(scale) / np.log(n))
            if fontsize_max is not None:
                fontsize = min(fontsize_max, fontsize)

        # apply special scaling factor for this backend
        fontsize = int(round(fontsize * self._font_scale_factor))
        return Font(fontname, fontsize)

    def set_line(self, color, alpha=1.0, linewidth=1, linestyle='solid'):
        self.line = self.get_line(color=color, linewidth=linewidth,
                                  linestyle=linestyle, alpha=alpha)
        return self.line

    def set_fill(self, color, alpha=1.0):
        self.fill = self.get_fill(color=color, alpha=alpha)
        return self.fill

    def set_font(self, fontname, fontsize=None):
        self.font = self.get_font(fontname, fontsize=fontsize)
        return self.font

    def get_line_from_shape(self, shape):
        linewidth = getattr(shape, 'linewidth', 1.0)
        linestyle = getattr(shape, 'linestyle', 'solid')
        alpha = getattr(shape, 'alpha', 1.0)
        color = getattr(shape, 'color', 'black')
        return self.get_line(color, alpha=alpha,
                             linewidth=linewidth, linestyle=linestyle)

    def get_fill_from_shape(self, shape):
        fill = getattr(shape, 'fill', False)
        if fill:
            if hasattr(shape, 'fillcolor') and shape.fillcolor is not None:
                color = shape.fillcolor
            else:
                color = shape.color

            alpha = getattr(shape, 'alpha', None)
            fillalpha = getattr(shape, 'fillalpha', alpha)
            fill = self.get_fill(color, alpha=fillalpha)
        else:
            fill = None
        return fill

    def get_font_from_shape(self, shape):
        if hasattr(shape, 'font'):
            if (hasattr(shape, 'fontsize') and shape.fontsize is not None and
                not getattr(shape, 'fontscale', False)):
                # user specified a font size and no scaling
                fontsize = shape.fontsize
            else:
                # user specified to scale font based on zoom level
                # of viewer
                # TODO: I think we can use the scaling in get_font()
                fontsize = shape.scale_font(self.viewer)

            # render-specific scaling
            #fontsize = self.scale_fontsize(fontsize)
            font = self.get_font(shape.font, fontsize=fontsize)
        else:
            font = None
        return font

    def set_line_from_shape(self, shape):
        self.line = self.get_line_from_shape(shape)
        return self.line

    def set_fill_from_shape(self, shape):
        self.fill = self.get_fill_from_shape(shape)
        return self.fill

    def set_font_from_shape(self, shape):
        self.font = self.get_font_from_shape(shape)
        return self.font

    def initialize_from_shape(self, shape, line=False, fill=False, font=False):
        if line:
            self.set_line_from_shape(shape)
        if fill:
            self.set_fill_from_shape(shape)
        if font:
            self.set_font_from_shape(shape)

    def scale_fontsize(self, fontsize):
        # TO BE EVENTUALLY DEPRECATED
        return self.renderer.scale_fontsize(fontsize)

    ##### DRAWING OPERATIONS #####

    def draw_image(self, cvs_img, cpoints, cache, whence, order='RGBA'):
        pass

    def draw_text(self, cx, cy, text, rot_deg=0.0, font=None, fill=None,
                  line=None):
        pass

    def draw_polygon(self, cpoints, line=None, fill=None):
        pass

    def draw_circle(self, cx, cy, cradius, line=None, fill=None):
        pass

    def draw_line(self, cx1, cy1, cx2, cy2, line=None):
        pass

    def draw_path(self, cpoints, line=None):
        pass


class RendererBase(object):
    """Base class from which all Renderer classes are derived."""

    def __init__(self, viewer):
        self.viewer = viewer
        self.logger = viewer.get_logger()
        self.surface = None

    def initialize(self):
        #raise RenderError("subclass should override this method!")
        pass

    def render_whence(self, whence):
        """Called in eventual response to a redraw request in the viewer.

        Causes the rendering pipeline to be partially or completely
        re-executed, depending on `whence`.

        Parameters
        ----------
        whence : `float` or `int`
            A number indicating at what level things need to be redrawn.

        """
        raise RenderError("subclass should override this method!")

    def finalize(self):
        #raise RenderError("subclass should override this method!")
        pass

    def prepare_image(self, cvs_img, cache, whence):
        """Prepare an image to be displayed somewhere in the viewer.

        Called at a point in the rendering process when it is known
        that an image will need to be displayed.  In this renderer,
        it is called via the "overlays" stage, and it creates an RGB
        image in the cache which is then merged into the background.

        Parameters
        ----------
        cvs_img : instance of `~ginga.canvas.types.image.Image`
            The canvas object for the image to be displayed

        cache : `~ginga.misc.Bunch.Bunch`
            A cache of items used by the renderer for a particular viewer

        whence : `float` or `int`
            A number indicating at what level things need to be redrawn.
        """
        raise RenderError("subclass should override this method!")

    def getwin_array(self, order='RGBA', alpha=1.0, dtype=None):
        """Get Numpy data array representing the viewer window.

        Parameters
        ----------
        order : str
            The desired order of RGB color layers.

        alpha : float
            Opacity.

        dtype : numpy dtype
            Numpy data type desired; defaults to the renderer default.

        Returns
        -------
        outarr : ndarray
            Numpy data array for display window.

        """
        raise RenderError("subclass should override this method!")

    def get_surface_as_array(self, order=None):
        raise RenderError("subclass should override this method!")

    def get_surface_as_bytes(self, order=None):
        """Returns the surface area as a bytes encoded RGB image buffer.
        Subclass should override if there is a more efficient conversion
        than from generating a numpy array first.
        """
        arr8 = self.get_surface_as_array(order=order)
        return arr8.tobytes(order='C')

    def get_surface_as_buffer(self, output=None, order=None):
        obuf = output
        if obuf is None:
            obuf = BytesIO()

        obuf.write(self.get_surface_as_bytes(order=order))
        return obuf

    def get_surface_as_rgb_format_buffer(self, output=None, format='png',
                                         quality=90):
        if self.surface is None:
            raise RenderError("No surface defined")

        # Get current surface as an array
        # note: use format 'RGB' because PIL cannot write JPEG with
        #  4 band images
        arr8 = self.get_surface_as_array(order='RGB')

        obuf = output
        if obuf is None:
            obuf = BytesIO()

        # make a PIL image
        image = Image.fromarray(arr8)

        image.save(obuf, format=format, quality=quality)
        if output is not None:
            return None
        return obuf

    def get_surface_as_rgb_format_bytes(self, format='png', quality=90):
        buf = self.get_surface_as_rgb_format_buffer(format=format,
                                                    quality=quality)
        return buf.getvalue()

    def save_surface_as_rgb_format_file(self, filepath, format='png',
                                        quality=90):
        if self.surface is None:
            raise RenderError("No surface defined")

        with open(filepath, 'wb') as out_f:
            self.get_surface_as_rgb_format_buffer(output=out_f, format=format,
                                                  quality=quality)
        self.logger.debug("wrote %s file '%s'" % (format, filepath))

    def reorder(self, dst_order, arr, src_order=None):
        """Reorder the output array to match that needed by the viewer."""
        if dst_order is None:
            dst_order = self.viewer.rgb_order
        if src_order is None:
            src_order = self.rgb_order
        if src_order != dst_order:
            arr = trcalc.reorder_image(dst_order, arr, src_order)

        return arr

    def scale(self, scales):
        self.viewer.redraw(whence=0)

    def pan(self, pos):
        self.viewer.redraw(whence=0)

    def rotate_2d(self, rot_deg):
        self.viewer.redraw(whence=2.6)

    def transform_2d(self, state):
        self.viewer.redraw(whence=2.5)

    def rgbmap_change(self, rgbmap):
        self.viewer.redraw(whence=2)

    def levels_change(self, levels):
        self.viewer.redraw(whence=1)

    def bg_change(self, bg):
        self.viewer.redraw(whence=0)

    def fg_change(self, fg):
        self.viewer.redraw(whence=3)

    def icc_profile_change(self):
        self.viewer.redraw(whence=2.3)

    def interpolation_change(self, interp):
        self.viewer.redraw(whence=0)

    def limits_change(self, limits):
        self.viewer.redraw(whence=3)

    def get_scale(self):
        """Return the scale currently set in the renderer."""
        raise RenderError("subclass should override this method!")

    def get_origin(self):
        """Return the point in data coordinates that is supposed to
        be at the pan position (center of the window)."""
        raise RenderError("subclass should override this method!")

    def get_window_center(self):
        """Return the point in pixel coordinates at the center
        of the window."""
        raise RenderError("subclass should override this method!")

    def get_window_size(self):
        """Return the window dimensions."""
        raise RenderError("subclass should override this method!")

    def resize(self, dims):
        raise RenderError("subclass should override this method!")

    def calc_const_len(self, clen):
        # For standard pixel renderer, pixel size is constant
        raise RenderError("subclass should override this method!")

    def _confirm_pan_and_scale(self, scale_x, scale_y, pan_x, pan_y,
                               win_wd, win_ht):
        raise RenderError("subclass should override this method!")

    def scale_fontsize(self, fontsize):
        raise RenderError("subclass should override this method!")


def get_render_class(rtype):

    rtype = rtype.lower()
    if rtype == 'pil':
        from ginga.pilw import CanvasRenderPil
        return CanvasRenderPil.CanvasRenderer

    if rtype == 'agg':
        from ginga.aggw import CanvasRenderAgg
        return CanvasRenderAgg.CanvasRenderer

    if rtype == 'opencv':
        from ginga.cvw import CanvasRenderCv
        return CanvasRenderCv.CanvasRenderer

    if rtype == 'cairo':
        from ginga.cairow import CanvasRenderCairo
        return CanvasRenderCairo.CanvasRenderer

    if rtype == 'matplotlib':
        from ginga.mplw import CanvasRenderMpl
        return CanvasRenderMpl.CanvasRenderer

    if rtype == 'opengl':
        from ginga.opengl import CanvasRenderGL
        return CanvasRenderGL.CanvasRenderer

    if rtype == 'qt':
        from ginga.qtw import CanvasRenderQt
        return CanvasRenderQt.CanvasRenderer

    if rtype == 'vqt':
        from ginga.qtw import CanvasRenderQt
        return CanvasRenderQt.VectorCanvasRenderer

    raise ValueError("Don't know about '%s' renderer type" % (rtype))


class StandardPipelineRenderer(RendererBase):
    """Standard renderer for generating bitmap-based image that can be
    copied to an RGB image type-widget or a canvas.

    This renderer has a pipeline that looks like:

    [createbg] => [overlays] => [iccprof] => [flipswap] => [rotate] => [output]

    """

    def __init__(self, viewer):
        super(StandardPipelineRenderer, self).__init__(viewer)

        # stages making up the rendering pipeline
        self.stage = Bunch.Bunch(createbg=render.CreateBg(viewer),
                                 overlays=render.Overlays(viewer),
                                 iccprofile=render.ICCProf(viewer),
                                 flipswap=render.FlipSwap(viewer),
                                 rotate=render.Rotate(viewer),
                                 output=render.Output(viewer))
        # build the pipeline
        self.pipeline = pipeline.Pipeline(self.logger,
                                          [self.stage.createbg,
                                           self.stage.overlays,
                                           self.stage.iccprofile,
                                           self.stage.flipswap,
                                           self.stage.rotate,
                                           self.stage.output],
                                          name='standard-pipeline-renderer')

        # A table of (threshold, stage) tuples. See render_whence()
        self.tbl = [(2.0, self.stage.overlays),
                    (2.3, self.stage.iccprofile),
                    (2.5, self.stage.flipswap),
                    (2.6, self.stage.rotate),
                    ]

        # order of RGBA channels that renderer needs to work in
        self.rgb_order = 'RGBA'
        self.std_order = 'RGBA'
        self.dims = (0, 0)
        self.state = Bunch.Bunch(org_scale=(1.0, 1.0, 1.0),
                                 org_pan=(0.0, 0.0, 0.0),
                                 ctr=(0, 0),
                                 win_dim=(0, 0),
                                 order=self.std_order)
        self.pipeline.set(state=self.state)
        # initialize pipeline
        self.pipeline.invalidate()

    def get_rgb_order(self):
        return self.rgb_order

    def invalidate(self):
        self.pipeline.invalidate()

    def initialize(self):
        # not currently used by this renderer
        pass

    def render_whence(self, whence):
        if whence < 3.0:
            # Search the table and stop at the first stage whose threshold
            # is <= whence. Run the pipeline from that stage.
            for thr, stage in self.tbl:
                if whence <= thr:
                    self.pipeline.set(whence=whence)
                    self.pipeline.run_from(stage)
                    break

        # TODO: what are options for high bit depth under the backend?
        last_stage = self.pipeline[-1]
        rgbarr = self.pipeline.get_data(last_stage)

        # NOTE: we assume Output stage has rendered according to the
        # RGB order needed by the renderer
        dst_order = self.get_rgb_order()
        out_order = self.pipeline.get('out_order')
        if dst_order != out_order:
            raise RenderError("RGB order of pipeline output ({}) "
                              "does not match renderer expected ({})".format(
                                  out_order, dst_order))

        self.render_image(rgbarr, dst_order, (0, 0))

    def finalize(self):
        # not currently used by this renderer
        pass

    def prepare_image(self, cvs_img, cache, whence):
        stage = self.pipeline[1]
        if cvs_img.kind == 'image':
            stage._prepare_image(cvs_img, cache, whence)
        elif cvs_img.kind == 'normimage':
            stage._prepare_norm_image(cvs_img, cache, whence)
        else:
            raise RenderError("I don't know how to render canvas type '{}'".format(cvs_img.kind))

    def getwin_array(self, order='RGBA', alpha=1.0, dtype=None):
        # NOTE: alpha parameter temporarily ignored for now
        dst_order = order.upper()
        #src_order = self.std_order
        src_order = self.get_rgb_order()

        last_stage = self.pipeline[-1]
        outarr = self.pipeline.get_data(last_stage)

        # reorder as caller needs it
        outarr = self.reorder(dst_order, outarr, src_order=src_order)
        outarr = np.ascontiguousarray(outarr)
        if dtype is not None:
            outarr = outarr.astype(dtype, copy=False)
        return outarr

    def get_scale(self):
        return self.state.org_scale

    def get_origin(self):
        return self.state.org_pan

    def get_window_center(self):
        return self.state.ctr

    def get_window_size(self):
        return self.state.win_dim

    def resize(self, dims):
        self._resize(dims)
        #self.pipeline.invalidate()
        self.pipeline.run_stage_idx(0)
        self.viewer.redraw(whence=0)

    def calc_const_len(self, clen):
        # For standard pipeline renderer, pixel size is constant
        return clen

    def _resize(self, dims):
        self.dims = dims
        # update window size and center in pipeline
        wd, ht = dims[:2]
        ctr = (wd // 2, ht // 2)
        self.state.setvals(win_dim=dims[:2], ctr=ctr,
                           order=self.std_order)

        # update pan and scale values in pipeline
        pan_x, pan_y = self.viewer.get_pan(coord='data')[:2]
        scale_x, scale_y = self.viewer.get_scale_xy()
        self._update_pan_and_scale(scale_x, scale_y,
                                   pan_x, pan_y, wd, ht)

    def _update_pan_and_scale(self, scale_x, scale_y, pan_x, pan_y,
                              win_wd, win_ht):
        data_off = self.viewer.data_off

        # record location of pan position pixel
        org_x, org_y = pan_x - data_off, pan_y - data_off

        org_scale_x, org_scale_y = scale_x, scale_y
        org_scale_z = (scale_x + scale_y) * 0.5

        self.state.setvals(org_pan=(org_x, org_y, 0.0),
                           org_scale=(org_scale_x, org_scale_y, org_scale_z))

    def _confirm_pan_and_scale(self, scale_x, scale_y, pan_x, pan_y,
                               win_wd, win_ht):
        # Sanity check on the scale
        sx = float(win_wd) / scale_x
        sy = float(win_ht) / scale_y
        if (sx < 1.0) or (sy < 1.0):
            if self.viewer.settings.get('sanity_check_scale', True):
                raise RenderError("new scale would exceed pixel max; scale unchanged")

        self._update_pan_and_scale(scale_x, scale_y, pan_x, pan_y,
                                   win_wd, win_ht)

    def scale_fontsize(self, fontsize):
        return font_asst.scale_fontsize(self.kind, fontsize)


def get_color(color, alpha):
    if color is not None:
        r, g, b = colors.resolve_color(color)
    else:
        r, g, b, alpha = 1.0, 1.0, 1.0, 0.0
    return (r, g, b, alpha)



#
# GlHelp.py -- help classes for OpenGL drawing
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.

import os.path

from OpenGL import GL as gl

import ginga.fonts
from ginga.canvas import transform

# Set up known fonts
fontdir, xx = os.path.split(ginga.fonts.__file__)
known_font = os.path.join(fontdir, 'Roboto', 'Roboto-Regular.ttf')

font_cache = {}


def get_cached_font(fontpath, fontsize):
    global font_cache

    key = (fontpath, fontsize)
    try:
        return font_cache[key]

    except KeyError:
        from PIL import ImageFont

        # TODO: try to lookup font before overriding
        fontpath = known_font

        font = ImageFont.truetype(fontpath, fontsize)
        font_cache[key] = font
        return font


class GlContext(object):

    def __init__(self, widget):
        self.widget = widget

    def text_extents(self, text, font):
        # TODO: we need a better approximation
        wd = len(text) * font.fontsize * 0.45
        ht = font.fontsize
        return wd, ht


class ShaderManager:
    """Class for building/managing/using GLSL shader programs.
    """
    def __init__(self, logger):
        self.logger = logger

        self.pgms = {}
        self.program = None
        self.shader = None

    def build_program(self, name, vertex_source, fragment_source):
        """Build a GL shader program from vertex and fragment sources.

        Parameters
        ----------
        name : str
            Name under which to store this program

        vertex_source : str
            source code for the vertex shader

        fragment_source : str
            source code for the fragment shader

        Returns
        -------
        pgm_id : int
            program id of the compiled shader
        """
        pgm_id = gl.glCreateProgram()
        vert_id = self._add_shader(vertex_source, gl.GL_VERTEX_SHADER)
        frag_id = self._add_shader(fragment_source, gl.GL_FRAGMENT_SHADER)

        gl.glAttachShader(pgm_id, vert_id)
        gl.glAttachShader(pgm_id, frag_id)
        gl.glLinkProgram(pgm_id)

        if gl.glGetProgramiv(pgm_id, gl.GL_LINK_STATUS) != gl.GL_TRUE:
            info = gl.glGetProgramInfoLog(pgm_id)
            gl.glDeleteProgram(pgm_id)
            gl.glDeleteShader(vert_id)
            gl.glDeleteShader(frag_id)
            self.logger.error('Error linking GLSL program: %s' % (info))
            raise RuntimeError('Error linking GLSL program: %s' % (info))

        gl.glDeleteShader(vert_id)
        gl.glDeleteShader(frag_id)

        self.pgms[name] = pgm_id
        return pgm_id

    def load_program(self, name, dirpath):
        """Load a GL shader program from sources on disk.

        Parameters
        ----------
        name : str
            Name under which to store this program

        dirpath : str
            path to where the vertex and fragment shader sources are stored

        Returns
        -------
        pgm_id : int
            program id of the compiled shader
        """
        vspath = os.path.join(dirpath, name + '.vert')
        with open(vspath, 'r') as in_f:
            vert_source = in_f.read().encode()

        fgpath = os.path.join(dirpath, name + '.frag')
        with open(fgpath, 'r') as in_f:
            frag_source = in_f.read().encode()

        return self.build_program(name, vert_source, frag_source)

    def setup_program(self, name):
        """Set up to use a shader program.

        Parameters
        ----------
        name : str
            Name of the shader program to use

        Returns
        -------
        shader :
            The OpenGL shader program
        """
        self.program = name
        if name is None:
            gl.glUseProgram(0)
            self.shader = None
        else:
            self.shader = self.pgms[name]
            gl.glUseProgram(self.shader)
        return self.shader

    def get_uniform_loc(self, attr_name):
        """Get the location of a shader program uniform variable.

        Parameters
        ----------
        attr_name : str
            Name of the shader program attribute

        Returns
        -------
        loc : int
            The location of the attribute
        """
        _loc = gl.glGetUniformLocation(self.shader, attr_name)
        return _loc

    def _add_shader(self, source, shader_type):
        try:
            shader_id = gl.glCreateShader(shader_type)
            gl.glShaderSource(shader_id, source)
            gl.glCompileShader(shader_id)
            if gl.glGetShaderiv(shader_id, gl.GL_COMPILE_STATUS) != gl.GL_TRUE:
                info = gl.glGetShaderInfoLog(shader_id)
                raise RuntimeError('Shader compilation failed: %s' % (info))

            return shader_id

        except Exception as e:
            gl.glDeleteShader(shader_id)
            raise


def get_transforms(v):
    tform = {
        'window_to_native': (transform.CartesianWindowTransform(v).invert() +
                             transform.RotationTransform(v).invert() +
                             transform.ScaleTransform(v).invert()),
        'cartesian_to_window': (transform.FlipSwapTransform(v) +
                                transform.CartesianWindowTransform(v)),
        'cartesian_to_native': (transform.FlipSwapTransform(v) +
                                transform.RotationTransform(v) +
                                transform.CartesianNativeTransform(v)),
        'data_to_cartesian': (transform.DataCartesianTransform(v) +
                              transform.ScaleTransform(v)),
        'data_to_scrollbar': (transform.DataCartesianTransform(v) +
                              transform.FlipSwapTransform(v) +
                              transform.RotationTransform(v)),
        'mouse_to_data': (
            transform.InvertedTransform(transform.DataCartesianTransform(v) +
                                        transform.ScaleTransform(v) +
                                        transform.FlipSwapTransform(v) +
                                        transform.RotationTransform(v) +
                                        transform.CartesianWindowTransform(v)
                                        )),
        'data_to_window': (transform.DataCartesianTransform(v) +
                           transform.ScaleTransform(v) +
                           transform.FlipSwapTransform(v) +
                           transform.RotationTransform(v) +
                           transform.CartesianWindowTransform(v)
                           ),
        'data_to_percentage': (transform.DataCartesianTransform(v) +
                               transform.ScaleTransform(v) +
                               transform.FlipSwapTransform(v) +
                               transform.RotationTransform(v) +
                               transform.CartesianWindowTransform(v) +
                               transform.WindowPercentageTransform(v)),
        'data_to_native': (transform.DataCartesianTransform(v) +
                           transform.FlipSwapTransform(v)
                           ),
        'wcs_to_data': transform.WCSDataTransform(v),
        'wcs_to_native': (transform.WCSDataTransform(v) +
                          transform.DataCartesianTransform(v) +
                          transform.FlipSwapTransform(v)),
    }
    return tform



# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
"""
The ``LoaderConfig`` plugin allows you to configure the file openers that
can be used to load various content into Ginga.

Registered file openers are associated with file MIME types, and there can
be several openers for a single MIME type.  A priority associated
with a MIME type/opener pairing determines which opener will be used
for each type--the lowest priority value will determine which opener will
be used.  If there are more than one opener with the same low priority
then the user will be prompted for which opener to use, when opening a
file in Ginga.  This plugin can be used to set the opener preferences
and save it to the user's $HOME/.ginga configuration area.

**Plugin Type: Global**

``LoaderConfig`` is a global plugin.  Only one instance can be opened.

**Usage**

After starting the plugin, the display will show all the registered MIME
types and the openers registered for those types, with an associated
priority for each MIME type/opener pairing.

Select one or more lines and type a priority for them in the box labeled
"Priority:"; press "Set" (or ENTER) to set the priority of those items.

.. note:: The lower the number, the higher the priority. Negative numbers
          are fine and the default priority for a loader is usually 0.
          So, for example, if there are two loaders available for a MIME
          type and one priority is set to -1 and the other to 0, the one
          with -1 will be used without asking the user to choose.


Click "Save" to save the priorities to $HOME/.ginga/loaders.json so that
they will be reloaded and used on subsequent restarts of the program.
"""
import os.path

import yaml

from ginga import GingaPlugin
from ginga.util.paths import ginga_home
from ginga.util import loader
from ginga.gw import Widgets

__all__ = ['LoaderConfig']


class LoaderConfig(GingaPlugin.GlobalPlugin):

    def __init__(self, fv):
        super().__init__(fv)

        self.loader_dct = dict()

        self.columns = [("Name", 'name'),
                        ("Priority", 'priority'),
                        ("Note", 'note')]

        self.gui_up = False

    def build_gui(self, container):
        vbox = Widgets.VBox()
        vbox.set_spacing(1)

        tv = Widgets.TreeView(sortable=True, use_alt_row_color=True,
                              selection='multiple', auto_expand=True)
        tv.add_callback('selected', self.select_cb)
        tv.setup_table(self.columns, 2, 'name')
        self.w.loader_tbl = tv

        vbox.add_widget(tv, stretch=1)

        tbar = Widgets.HBox()
        tbar.set_border_width(4)
        tbar.set_spacing(4)

        tbar.add_widget(Widgets.Label('Priority:'))
        pri = Widgets.TextEntrySet(editable=True)
        pri.set_tooltip("Edit priority of loader (lower=better, negative numbers ok)")
        pri.set_enabled(False)
        pri.add_callback('activated', self.set_priority_cb)
        self.w.pri_edit = pri
        tbar.add_widget(pri)

        tbar.add_widget(Widgets.Label(''), stretch=1)
        vbox.add_widget(tbar, stretch=0)

        btns = Widgets.HBox()
        btns.set_border_width(4)
        btns.set_spacing(4)

        btn = Widgets.Button("Close")
        btn.add_callback('activated', lambda w: self.close())
        btns.add_widget(btn)
        btn = Widgets.Button("Help")
        btn.add_callback('activated', lambda w: self.help())
        btns.add_widget(btn, stretch=0)
        btn = Widgets.Button("Save")
        btn.add_callback('activated', lambda w: self.save_loaders_cb())
        btn.set_tooltip("Save configuration of loaders")
        btns.add_widget(btn, stretch=0)

        btns.add_widget(Widgets.Label(''), stretch=1)
        vbox.add_widget(btns, stretch=0)
        container.add_widget(vbox, stretch=1)

        self.gui_up = True

    def start(self):
        # create loader table
        tree_dct = dict()
        for mimetype, dct in loader.loader_by_mimetype.items():
            md = dict()
            tree_dct[mimetype] = md
            for name, bnch in dct.items():
                md[name] = dict(name=bnch.opener.name,
                                priority=bnch.priority,
                                note=bnch.opener.note)
            self.loader_dct = tree_dct

        self.w.loader_tbl.set_tree(self.loader_dct)
        self.w.loader_tbl.set_optimal_column_widths()

    def stop(self):
        self.gui_up = False

    def set_priority_cb(self, w):
        sel_dct = self.w.loader_tbl.get_selected()
        priority = int(self.w.pri_edit.get_text())
        for mimetype, ld_dct in sel_dct.items():
            for name, m_dct in ld_dct.items():
                self.loader_dct[mimetype][name]['priority'] = priority
                # actually change it in the loader registration
                opener = loader.get_opener(name)
                loader.add_opener(opener, [mimetype], priority=priority,
                                  note=opener.__doc__)
        # update the UI table
        self.w.loader_tbl.set_tree(self.loader_dct)

    def save_loaders_cb(self):
        path = os.path.join(ginga_home, 'loaders.yml')
        try:
            with open(path, 'w') as out_f:
                out_f.write(yaml.dump(self.loader_dct, indent=4))

        except Exception as e:
            self.logger.error(f"failed to save loader file: {e}",
                              exc_info=True)
            self.fv.show_error(str(e))

    def select_cb(self, w, dct):
        selected = len(dct) > 0
        self.w.pri_edit.set_enabled(selected)
        if not selected:
            self.w.pri_edit.set_text('')
        else:
            # only set priority widget if all selected are the same
            # unique value
            priorities = set([l_dct['priority']
                              for m_dct in dct.values()
                              for l_dct in m_dct.values()])
            if len(priorities) == 1:
                self.w.pri_edit.set_text(str(priorities.pop()))
            else:
                self.w.pri_edit.set_text('')

    def close(self):
        self.fv.stop_global_plugin(str(self))
        return True

    def __str__(self):
        return 'loaderconfig'



# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
"""
The ``Info`` plugin provides a pane of commonly useful metadata about the
focused channel image.  Common information includes some metadata
header values, coordinates, dimensions of the image, minimum and
maximum values, etc.  As the cursor is moved around the image, the X, Y,
Value, RA, and DEC values are updated to reflect the value under the cursor.

**Plugin Type: Global**

``Info`` is a global plugin.  Only one instance can be opened.

**Usage**

At the bottom of the ``Info`` interface are the color distribution
and cut levels controls.  The selector above the cut levels boxes lets
you chose from several distribution algorithms that map the values in
the image to the color map.  Choices are "linear", "log", "power", "sqrt",
"squared", "asinh", "sinh", and "histeq" (histogram equalization).

Below this, the low and high cut levels are shown and can be adjusted.
Pressing the "Auto Levels" button will recalculate cut levels based on
the current auto cut levels algorithm and parameters defined in the channel
preferences.

Below the "Auto Levels" button, the status of the settings for
"Cut New", "Zoom New", and "Center New" are shown for the currently active
channel.  These indicate how new images that are added to the channel
will be affected by auto cut levels, fitting to the window and panning
to the center of the image.

The "Follow New" checkbox controls whether the viewer will automatically
display new images added to the channel.  The "Raise New" checkbox controls
whether an image viewer window is raised when a new image is added.  These
two controls can be useful, for example, if an external program is adding
images to the viewer, and you wish to prevent interruption of your work
examining a particular image.

As a global plugin, ``Info`` responds to a change of focus to a new channel
by displaying the metadata from the new channel.
It typically appears under the "Synopsis" tab in the user interface.

This plugin is not usually configured to be closeable, but the user can
make it so by setting the "closeable" setting to True in the configuration
file--then Close and Help buttons will be added to the bottom of the UI.

"""
import time
import numpy as np

from ginga.gw import Widgets
from ginga import GingaPlugin, ColorDist
from ginga.rv.plugins.Toolbar import Toolbar
from ginga.table.AstroTable import AstroTable
from ginga.plot.Plotable import Plotable
from ginga.canvas.CanvasObject import get_canvas_types

__all__ = ['Info', 'Info_Ginga_Image',
           'Info_Ginga_Plot', 'Info_Ginga_Table']


class Info(Toolbar):

    def __init__(self, fv):
        # superclass defines some variables for us, like logger
        super().__init__(fv)

        self.opname_prefix = 'Info_'

    def __str__(self):
        return 'info'


class Info_Common(GingaPlugin.LocalPlugin):

    def __init__(self, fv, chviewer):
        # superclass defines some variables for us, like logger
        super().__init__(fv, chviewer)

        # truncate names after this size
        self.maxstr = 60

    def trunc(self, s):
        if len(s) > self.maxstr:
            return s[:self.maxstr - 3] + '...'
        else:
            return s


class Info_Ginga_Image(Info_Common):
    """Info sidebar for the Ginga Image viewer.
    """

    def __init__(self, fv, chviewer):
        # superclass defines some variables for us, like logger
        super().__init__(fv, chviewer)

        self.autozoom_options = ['on', 'override', 'once', 'off']
        self.autocut_options = self.autozoom_options
        self.autocenter_options = self.autozoom_options

        self.fv.add_callback('field-info', self.field_info_cb)

        self.gui_up = False

    def build_gui(self, container):
        sw = Widgets.ScrollArea()

        vbox = Widgets.VBox()
        sw.set_widget(vbox)

        captions = (('Channel:', 'label', 'Channel', 'llabel'),
                    ('Name:', 'label', 'Name', 'llabel'),
                    ('Object:', 'label', 'Object', 'llabel'),
                    ('X:', 'label', 'X', 'llabel'),
                    ('Y:', 'label', 'Y', 'llabel'),
                    ('Image-X:', 'label', 'Image_X', 'llabel'),
                    ('Image-Y:', 'label', 'Image_Y', 'llabel'),
                    ('Value:', 'label', 'Value', 'llabel'),
                    ('RA:', 'label', 'RA', 'llabel'),
                    ('DEC:', 'label', 'DEC', 'llabel'),
                    ('Equinox:', 'label', 'Equinox', 'llabel'),
                    ('Dimensions:', 'label', 'Dimensions', 'llabel'),
                    ('Min:', 'label', 'Min', 'llabel'),
                    ('Max:', 'label', 'Max', 'llabel'),
                    )
        w, b = Widgets.build_info(captions)
        self.w = b

        col = Widgets.VBox()
        row = Widgets.HBox()
        row.set_spacing(0)
        row.set_border_width(0)
        row.add_widget(w, stretch=0)
        row.add_widget(Widgets.Label(''), stretch=1)
        col.add_widget(row, stretch=0)
        col.add_widget(Widgets.Label(''), stretch=1)
        sw2 = Widgets.ScrollArea()
        sw2.set_widget(col)
        vbox.add_widget(sw2, stretch=1)

        captions = (('Zoom:', 'label', 'Zoom', 'llabel',
                     'Color Dist', 'combobox'),
                    ('Cut Low:', 'label', 'Cut Low Value', 'llabel',
                     'Cut Low', 'entry'),
                    ('Cut High:', 'label', 'Cut High Value', 'llabel',
                     'Cut High', 'entry'),
                    ('Auto Levels', 'button', 'spacer_1', 'spacer',
                     'Cut Levels', 'button'),
                    ('Cut New:', 'label', 'Cut New', 'combobox'),
                    ('Zoom New:', 'label', 'Zoom New', 'combobox',
                     'Follow New', 'checkbutton'),
                    ('Center New:', 'label', 'Center New', 'combobox',
                     'Raise New', 'checkbutton'),
                    )

        w, b = Widgets.build_info(captions)
        self.w.update(b)
        b.cut_levels.set_tooltip("Set cut levels manually")
        loval, hival = self.fitsimage.get_cut_levels()
        b.auto_levels.set_tooltip("Set cut levels by algorithm")
        b.cut_low.set_tooltip("Set low cut level (press Enter)")
        b.cut_low.set_length(9)
        b.cut_low.set_text(str(loval))
        b.cut_low_value.set_text(str(loval))
        b.cut_high.set_tooltip("Set high cut level (press Enter)")
        b.cut_high.set_length(9)
        b.cut_high.set_text(str(hival))
        b.cut_high_value.set_text(str(hival))

        combobox = b.color_dist
        for name in ColorDist.get_dist_names():
            combobox.append_text(name)
        b.color_dist.set_tooltip("Set distribution (stretching) algorithm")

        combobox = b.cut_new
        for name in self.autocut_options:
            combobox.append_text(name)
        b.cut_new.set_tooltip("Automatically set cut levels when switching images")

        combobox = b.zoom_new
        for name in self.autozoom_options:
            combobox.append_text(name)
        b.zoom_new.set_tooltip("Automatically fit image to window when switching images")

        combobox = b.center_new
        for name in self.autocenter_options:
            combobox.append_text(name)
        b.center_new.set_tooltip("Automatically center image in window when switching images")

        b.follow_new.set_tooltip("Automatically switch to new images in channel")
        b.raise_new.set_tooltip("Automatically raise channel viewer for new images")

        row = Widgets.HBox()
        row.set_spacing(0)
        row.set_border_width(0)
        row.add_widget(w, stretch=0)
        row.add_widget(Widgets.Label(''), stretch=1)
        vbox.add_widget(row, stretch=0)

        self.w.channel.set_text(self.channel.name)
        self.w.color_dist.add_callback('activated', self.set_color_dist)
        self.w.cut_low.add_callback('activated', self.cut_levels)
        self.w.cut_high.add_callback('activated', self.cut_levels)
        self.w.cut_levels.add_callback('activated', self.cut_levels)
        self.w.auto_levels.add_callback('activated', self.auto_levels)
        self.w.cut_new.add_callback('activated', self.set_autocuts_cb)
        self.w.zoom_new.add_callback('activated', self.set_autozoom_cb)
        self.w.center_new.add_callback('activated', self.set_autocenter_cb)
        self.w.follow_new.add_callback('activated', self.set_follow_cb)
        self.w.raise_new.add_callback('activated', self.set_raise_cb)

        fitssettings = self.fitsimage.get_settings()
        for name in ['cuts']:
            fitssettings.get_setting(name).add_callback('set',
                                                        self.cutset_cb)
        for name in ['scale']:
            fitssettings.get_setting(name).add_callback('set',
                                                        self.zoomset_cb)
        fitssettings.get_setting('color_algorithm').add_callback('set',
                                                                 self.cdistset_cb)
        fitssettings.get_setting('autocuts').add_callback('set',
                                                          self.autocuts_cb)
        fitssettings.get_setting('autozoom').add_callback('set',
                                                          self.autozoom_cb)
        fitssettings.get_setting('autocenter').add_callback('set',
                                                            self.autocenter_cb)
        fitssettings.get_setting('switchnew').add_callback('set',
                                                           self.follow_cb)
        fitssettings.get_setting('raisenew').add_callback('set',
                                                          self.raise_cb)

        container.add_widget(sw, stretch=1)
        self.gui_up = True

    def start(self):
        self.redo()

    def stop(self):
        self.gui_up = False

    def close(self):
        # NOTE: this shouldn't be called under normal usage
        self.fv.stop_local_plugin(self.chname, str(self))

    def redo(self):
        self.set_info()

    def zoomset_cb(self, setting, value):
        """This callback is called when the main window is zoomed.
        """
        if not self.gui_up:
            return
        #scale_x, scale_y = self.fitsimage.get_scale_xy()
        scale_x, scale_y = value

        # Set text showing zoom factor (1X, 2X, etc.)
        if scale_x == scale_y:
            text = self.fv.scale2text(scale_x)
        else:
            textx = self.fv.scale2text(scale_x)
            texty = self.fv.scale2text(scale_y)
            text = "X: %s  Y: %s" % (textx, texty)
        self.w.zoom.set_text(text)

    def cutset_cb(self, setting, value):
        if not self.gui_up:
            return
        loval, hival = value
        #self.w.cut_low.set_text('%.4g' % (loval))
        self.w.cut_low_value.set_text('%.4g' % (loval))
        #self.w.cut_high.set_text('%.4g' % (hival))
        self.w.cut_high_value.set_text('%.4g' % (hival))

    def cdistset_cb(self, setting, value):
        if not self.gui_up:
            return
        name = value
        self.w.color_dist.set_text(name)

    def autocuts_cb(self, setting, option):
        if not self.gui_up:
            return
        self.logger.debug("autocuts changed to %s" % option)
        index = self.autocut_options.index(option)
        self.w.cut_new.set_index(index)

    def autozoom_cb(self, setting, option):
        if not self.gui_up:
            return
        index = self.autozoom_options.index(option)
        self.w.zoom_new.set_index(index)

    def autocenter_cb(self, setting, option):
        if not self.gui_up:
            return
        # Hack to convert old values that used to be T/F
        if isinstance(option, bool):
            choice = {True: 'on', False: 'off'}
            option = choice[option]
        index = self.autocenter_options.index(option)
        self.w.center_new.set_index(index)

    def follow_cb(self, setting, option):
        if not self.gui_up:
            return
        self.w.follow_new.set_state(option)

    def raise_cb(self, setting, option):
        if not self.gui_up:
            return
        self.w.raise_new.set_state(option)

    def set_autocuts_cb(self, w, index):
        if not self.gui_up:
            return
        option = self.autocut_options[index]
        self.fitsimage.enable_autocuts(option)

    def set_autozoom_cb(self, w, index):
        if not self.gui_up:
            return
        option = self.autozoom_options[index]
        self.fitsimage.enable_autozoom(option)

    def set_autocenter_cb(self, w, index):
        if not self.gui_up:
            return
        option = self.autocenter_options[index]
        self.fitsimage.enable_autocenter(option)

    def set_follow_cb(self, w, tf):
        if not self.gui_up:
            return
        self.fitsimage.get_settings().set(switchnew=tf)

    def set_raise_cb(self, w, tf):
        if not self.gui_up:
            return
        self.fitsimage.get_settings().set(raisenew=tf)

    # LOGIC

    def set_info(self):
        if not self.gui_up:
            return
        # Show cut levels
        loval, hival = self.fitsimage.get_cut_levels()
        #self.w.cut_low.set_text('%.4g' % (loval))
        self.w.cut_low_value.set_text('%.4g' % (loval))
        #self.w.cut_high.set_text('%.4g' % (hival))
        self.w.cut_high_value.set_text('%.4g' % (hival))

        # update zoom indicator
        scalefactor = self.fitsimage.get_scale()
        text = self.fv.scale2text(scalefactor)
        self.w.zoom.set_text(text)

        # update cut new/zoom new indicators
        t_ = self.fitsimage.get_settings()
        index = self.autocut_options.index(t_['autocuts'])
        self.w.cut_new.set_index(index)
        index = self.autozoom_options.index(t_['autozoom'])
        self.w.zoom_new.set_index(index)
        option = t_['autocenter']
        # Hack to convert old values that used to be T/F
        if isinstance(option, bool):
            choice = {True: 'on', False: 'off'}
            option = choice[option]
        index = self.autocenter_options.index(option)
        self.w.center_new.set_index(index)
        self.w.follow_new.set_state(t_['switchnew'])
        self.w.raise_new.set_state(t_['raisenew'])

        # Set color distribution indicator
        name = t_['color_algorithm']
        self.w.color_dist.set_text(name)

        image = self.fitsimage.get_image()
        if image is None:
            return
        header = image.get_header()

        # Update info panel
        name = self.trunc(image.get('name', 'Noname'))
        self.w.name.set_text(name)
        objtext = self.trunc(header.get('OBJECT', 'UNKNOWN'))
        self.w.object.set_text(objtext)
        equinox = header.get('EQUINOX', '')
        self.w.equinox.set_text(str(equinox))

        # Show min, max values
        width, height = self.fitsimage.get_data_size()
        minval, maxval = image.get_minmax(noinf=False)
        self.w.max.set_text(str(maxval))
        self.w.min.set_text(str(minval))

        # Show dimensions
        dim_txt = "%dx%d" % (width, height)
        self.w.dimensions.set_text(dim_txt)

    def field_info_cb(self, fv, channel, info):
        if not self.gui_up:
            return
        self.w.x.set_text("%.3f" % info.x)
        self.w.y.set_text("%.3f" % info.y)
        if 'image_x' in info:
            self.w.image_x.set_text("%.3f" % info.image_x)
        else:
            self.w.image_x.set_text("")
        if 'image_y' in info:
            self.w.image_y.set_text("%.3f" % info.image_y)
        else:
            self.w.image_y.set_text("")
        self.w.value.set_text(str(info.value))
        if 'ra_txt' in info:
            self.w.ra.set_text(info.ra_txt)
            self.w.dec.set_text(info.dec_txt)
        if 'ra_lbl' in info:
            self.w.lbl_ra.set_text(info.ra_lbl + ':')
            self.w.lbl_dec.set_text(info.dec_lbl + ':')

    def cut_levels(self, w):
        loval, hival = self.fitsimage.get_cut_levels()
        try:
            lostr = self.w.cut_low.get_text().strip()
            if lostr != '':
                loval = float(lostr)

            histr = self.w.cut_high.get_text().strip()
            if histr != '':
                hival = float(histr)
            self.logger.debug("locut=%f hicut=%f" % (loval, hival))

            return self.fitsimage.cut_levels(loval, hival)
        except Exception as e:
            self.fv.show_error("Error cutting levels: %s" % (str(e)))

        return True

    def auto_levels(self, w):
        self.fitsimage.auto_levels()

    def set_color_dist(self, w, idx):
        name = w.get_text()
        self.fitsimage.set_color_algorithm(name)

    def __str__(self):
        return 'info_ginga_image'


class Info_Ginga_Plot(Info_Common):
    """Info sidebar for the Ginga Plot viewer.
    """

    def __init__(self, fv, fitsimage):
        # superclass defines some variables for us, like logger
        super().__init__(fv, fitsimage)

        # Plugin preferences
        prefs = self.fv.get_preferences()
        self.settings = prefs.create_category('info_Ginga_Plot')
        self.settings.add_defaults(linewidth=1,
                                   linestyle='-',
                                   linecolor='blue',
                                   markersize=6,
                                   markerwidth=0.5,
                                   markerstyle='o',
                                   markercolor='red',
                                   show_marker=False,
                                   file_suffix='.png')
        self.settings.load(onError='silent')

        viewer = self.channel.get_viewer('Ginga Plot')
        self.plot_viewer = viewer

        viewer.add_callback('range-set', self.range_set_cb)
        viewer.add_callback('motion', self.motion_cb)

        self.gui_up = False

    def build_gui(self, container):
        # if not have_mpl:
        #     raise ImportError('Install matplotlib to use this plugin')
        t_ = self.plot_viewer.get_settings()
        sw = Widgets.ScrollArea()

        vbox = Widgets.VBox()
        sw.set_widget(vbox)

        captions = (('Channel:', 'label', 'channel', 'llabel'),
                    ('Name:', 'label', 'name', 'llabel'),
                    ('X:', 'label', 'x_val', 'llabel'),
                    ('Y:', 'label', 'y_val', 'llabel'),
                    )
        w, b = Widgets.build_info(captions)
        self.w = b

        self.w.channel.set_text(self.channel.name)

        col = Widgets.VBox()
        row = Widgets.HBox()
        row.set_spacing(0)
        row.set_border_width(0)
        row.add_widget(w, stretch=0)
        row.add_widget(Widgets.Label(''), stretch=1)
        col.add_widget(row, stretch=0)
        col.add_widget(Widgets.Label(''), stretch=1)
        sw2 = Widgets.ScrollArea()
        sw2.set_widget(col)
        vbox.add_widget(sw2, stretch=1)

        captions = (('spacer_1', 'spacer', 'X', 'llabel', 'Y', 'llabel'),
                    ('Dist:', 'label', 'x_dist', 'combobox',
                     'y_dist', 'combobox'),
                    ('Low:', 'label', 'x_lo', 'entry', 'y_lo', 'entry'),
                    ('High:', 'label', 'x_hi', 'entry', 'y_hi', 'entry'),
                    ('spacer_3', 'spacer', 'Show marker', 'checkbox',
                     'Save Plot', 'button'),
                    )
        w, b = Widgets.build_info(captions)
        self.w.update(b)

        # Controls for X-axis scaling
        combobox = b.x_dist
        for name in ['linear', 'log']:
            combobox.append_text(name)
        combobox.set_tooltip('Select a mapping to plot on X-axis')
        combobox.set_text(t_['plot_dist_axis'][0])
        combobox.add_callback('activated', self.x_dist_cb)

        # Controls for Y-axis column listing
        combobox = b.y_dist
        for name in ['linear', 'log']:
            combobox.append_text(name)
        combobox.set_tooltip('Select a mapping to plot on Y-axis')
        combobox.set_text(t_['plot_dist_axis'][1])
        combobox.add_callback('activated', self.y_dist_cb)

        self.set_xlimits_widgets()
        self.set_ylimits_widgets()

        b.x_lo.add_callback('activated', lambda w: self.set_xlim_lo_cb())
        b.x_lo.set_tooltip('Set X lower limit')

        b.x_hi.add_callback('activated', lambda w: self.set_xlim_hi_cb())
        b.x_hi.set_tooltip('Set X upper limit')

        b.y_lo.add_callback('activated', lambda w: self.set_ylim_lo_cb())
        b.y_lo.set_tooltip('Set Y lower limit')

        b.y_hi.add_callback('activated', lambda w: self.set_ylim_hi_cb())
        b.y_hi.set_tooltip('Set Y upper limit')

        b.show_marker.set_state(t_.get('plot_show_marker', False))
        b.show_marker.add_callback('activated', self.set_marker_cb)
        b.show_marker.set_tooltip('Mark data points')

        # Button to save plot
        b.save_plot.set_tooltip("Save plot to file")
        b.save_plot.add_callback('activated', lambda w: self.save_cb())
        b.save_plot.set_enabled(True)

        row = Widgets.HBox()
        row.set_spacing(0)
        row.set_border_width(0)
        row.add_widget(w, stretch=0)
        row.add_widget(Widgets.Label(''), stretch=1)
        vbox.add_widget(row, stretch=0)

        container.add_widget(sw, stretch=1)
        self.gui_up = True

    def redo(self):
        """This is called when a new image arrives or the data in the
        existing image changes.
        """
        if not self.gui_up:
            return

        dataobj = self.plot_viewer.get_dataobj()

        name = self.trunc(dataobj.get('name', 'Noname'))
        self.w.name.set_text(name)

    def set_xlimits_widgets(self, set_min=True, set_max=True):
        """Populate axis limits GUI with current plot values."""
        ranges = self.plot_viewer.get_ranges()
        (x_lo, x_hi), (y_lo, y_hi) = ranges
        if set_min:
            self.w.x_lo.set_text('{0}'.format(x_lo))
        if set_max:
            self.w.x_hi.set_text('{0}'.format(x_hi))

    def set_ylimits_widgets(self, set_min=True, set_max=True):
        """Populate axis limits GUI with current plot values."""
        ranges = self.plot_viewer.get_ranges()
        (x_lo, x_hi), (y_lo, y_hi) = ranges
        if set_min:
            self.w.y_lo.set_text('{0}'.format(y_lo))
        if set_max:
            self.w.y_hi.set_text('{0}'.format(y_hi))

    def x_dist_cb(self, w, index):
        _dist = w.get_text().lower()
        self.plot_viewer.set_dist_axis(x_axis=_dist)

    def y_dist_cb(self, w, index):
        _dist = w.get_text().lower()
        self.plot_viewer.set_dist_axis(y_axis=_dist)

    def set_xlim_lo_cb(self, redraw=True):
        """Set plot limit based on user values."""
        ranges = self.plot_viewer.get_ranges()
        (x_lo, x_hi), (y_lo, y_hi) = ranges

        try:
            x_min = float(self.w.x_lo.get_text())
        except Exception as e:
            self.fv.show_error(f"error setting X low limit: {e}")
            return

        self.plot_viewer.set_ranges(x_range=(x_min, x_hi))

    def set_xlim_hi_cb(self, redraw=True):
        """Set plot limit based on user values."""
        ranges = self.plot_viewer.get_ranges()
        (x_lo, x_hi), (y_lo, y_hi) = ranges

        try:
            x_max = float(self.w.x_hi.get_text())
        except Exception as e:
            self.fv.show_error(f"error setting X high limit: {e}")
            return

        self.plot_viewer.set_ranges(x_range=(x_lo, x_max))

    def set_ylim_lo_cb(self, redraw=True):
        """Set plot limit based on user values."""
        ranges = self.plot_viewer.get_ranges()
        (x_lo, x_hi), (y_lo, y_hi) = ranges

        try:
            y_min = float(self.w.y_lo.get_text())
        except Exception as e:
            self.fv.show_error(f"error setting Y low limit: {e}")
            return

        self.plot_viewer.set_ranges(y_range=(y_min, y_hi))

    def set_ylim_hi_cb(self, redraw=True):
        """Set plot limit based on user values."""
        ranges = self.plot_viewer.get_ranges()
        (x_lo, x_hi), (y_lo, y_hi) = ranges

        try:
            y_max = float(self.w.y_hi.get_text())
        except Exception as e:
            self.fv.show_error(f"error setting Y high limit: {e}")
            return

        self.plot_viewer.set_ranges(y_range=(y_lo, y_max))

    def range_set_cb(self, viewer, ranges):
        (xmin, xmax), (ymin, ymax) = ranges

        if self.gui_up:
            self.set_xlimits_widgets()
            self.set_ylimits_widgets()

    def motion_cb(self, viewer, event):
        if not self.gui_up:
            return
        data_x, data_y = event.data_x, event.data_y
        x_str = str(data_x)
        y_str = str(data_y)
        self.w.x_val.set_text(x_str)
        self.w.y_val.set_text(y_str)

    def set_marker_cb(self, w, tf):
        """Toggle show/hide data point markers."""
        settings = self.plot_viewer.get_settings()
        settings.set(plot_show_marker=tf)

    def save_cb(self):
        """Save plot to file."""

        # This just defines the basename.
        # Extension has to be explicitly defined or things can get messy.
        w = Widgets.SaveDialog(title='Save plot')
        target = w.get_path()
        if target is None:
            # Save canceled
            return

        plot_ext = self.settings.get('file_suffix', '.png')

        try:
            self.plot_viewer.save_rgb_image_as_file(target, format='png')

        except Exception as e:
            self.logger.error(str(e))
        else:
            self.logger.info('Table plot saved as {0}'.format(target))

    def start(self):
        self.redo()

    def stop(self):
        self.gui_up = False

    def close(self):
        # NOTE: this shouldn't be called under normal usage
        self.fv.stop_local_plugin(self.chname, str(self))

    def __str__(self):
        return 'info_ginga_plot'


class Info_Ginga_Table(Info_Common):
    """Info sidebar for the Ginga Table viewer.
    """

    def __init__(self, fv, chviewer):
        # superclass defines some variables for us, like logger
        super().__init__(fv, chviewer)

        viewer = self.channel.get_viewer('Ginga Table')
        self.table_viewer = viewer

        # To store all active table info
        self.tab = None
        self.cols = []
        self._idx = []
        self._idxname = '_idx'
        # To store selected columns names of active table
        self.x_col = ''
        self.y_col = ''

        self.gui_up = False

    def build_gui(self, container):
        sw = Widgets.ScrollArea()

        vbox = Widgets.VBox()
        sw.set_widget(vbox)

        captions = (('Channel:', 'label', 'channel', 'llabel'),
                    ('Name:', 'label', 'name', 'llabel'),
                    )
        w, b = Widgets.build_info(captions)
        self.w = b

        self.w.channel.set_text(self.channel.name)

        col = Widgets.VBox()
        row = Widgets.HBox()
        row.set_spacing(0)
        row.set_border_width(0)
        row.add_widget(w, stretch=0)
        row.add_widget(Widgets.Label(''), stretch=1)
        col.add_widget(row, stretch=0)
        col.add_widget(Widgets.Label(''), stretch=1)
        sw2 = Widgets.ScrollArea()
        sw2.set_widget(col)
        vbox.add_widget(sw2, stretch=1)

        channel = self.channel
        self.w.channel.set_text(channel.name)

        captions = (('spacer_1', 'spacer', 'X', 'llabel', 'Y', 'llabel'),
                    ('Col:', 'label', 'x_col', 'combobox', 'y_col', 'combobox'),
                    ('spacer_3', 'spacer', 'spacer_4', 'spacer',
                     'Make Plot', 'button'),
                    )
        w, b = Widgets.build_info(captions)
        self.w.update(b)

        # Controls for X-axis column listing
        combobox = b.x_col
        combobox.set_enabled(False)
        for name in self.cols:
            combobox.append_text(name)
        if self.x_col in self.cols:
            combobox.set_text(self.x_col)
        combobox.add_callback('activated', self.x_select_cb)
        combobox.set_tooltip('Select a column to plot on X-axis')

        # Controls for Y-axis column listing
        combobox = b.y_col
        combobox.set_enabled(False)
        for name in self.cols:
            combobox.append_text(name)
        if self.y_col in self.cols:
            combobox.set_text(self.y_col)
        combobox.add_callback('activated', self.y_select_cb)
        combobox.set_tooltip('Select a column to plot on Y-axis')

        # Button to save plot
        b.make_plot.set_tooltip("Plot selected columns")
        b.make_plot.add_callback('activated', self.make_plot_cb)
        b.make_plot.set_enabled(False)

        fr = Widgets.Frame("Plot Maker")
        fr.set_widget(w)
        row = Widgets.HBox()
        row.set_spacing(0)
        row.set_border_width(0)
        row.add_widget(fr, stretch=0)
        row.add_widget(Widgets.Label(''), stretch=1)
        vbox.add_widget(row, stretch=0)

        container.add_widget(sw, stretch=1)
        self.gui_up = True

    def start(self):
        self.redo()

    def stop(self):
        self.tab = None
        self.gui_up = False

    def close(self):
        # NOTE: this shouldn't be called under normal usage
        self.fv.stop_local_plugin(self.chname, str(self))

    def _set_combobox(self, combobox, vals, default=0):
        """Populate combobox with given list."""
        combobox.clear()
        for val in vals:
            combobox.append_text(val)
        if default > len(vals):
            default = 0
        val = vals[default]
        combobox.show_text(val)
        return val

    def setup_table(self, dataobj):
        # Generate column indices
        self.w.x_col.set_enabled(True)
        self.w.y_col.set_enabled(True)
        self.w.make_plot.set_enabled(True)
        # Generate column indices
        self.tab = dataobj.get_data()
        self._idx = np.arange(len(self.tab))

        # Populate combobox with table column names
        cols = [self._idxname] + self.tab.colnames
        if cols != self.cols:
            self.cols = cols
            self.x_col = self._set_combobox(self.w.x_col, self.cols, default=1)
            self.y_col = self._set_combobox(self.w.y_col, self.cols, default=2)

    def clear(self):
        self.tab = None
        self.cols = []
        self._idx = []
        self._idxname = '_idx'
        self.x_col = ''
        self.y_col = ''

    def redo(self):
        """This is called when a new image arrives or the data in the
        existing image changes.
        """
        if not self.gui_up:
            return

        dataobj = self.table_viewer.get_dataobj()
        if isinstance(dataobj, AstroTable):
            self.setup_table(dataobj)

        else:
            self.logger.info("not able to process this object")

        name = self.trunc(dataobj.get('name', 'Noname'))
        self.w.name.set_text(name)

    # LOGIC

    def _get_label(self, axis):
        """Return plot label for column for the given axis."""

        if axis == 'x':
            colname = self.x_col
        else:  # y
            colname = self.y_col

        if colname == self._idxname:
            label = 'Index'
        else:
            col = self.tab[colname]
            if col.unit:
                label = '{0} ({1})'.format(col.name, col.unit)
            else:
                label = col.name

        return label

    def x_select_cb(self, w, index):
        """Callback to set X-axis column."""
        try:
            self.x_col = self.cols[index]
        except IndexError as e:
            self.logger.error(str(e))

    def y_select_cb(self, w, index):
        """Callback to set Y-axis column."""
        try:
            self.y_col = self.cols[index]
        except IndexError as e:
            self.logger.error(str(e))

    def make_plot_cb(self, w):
        if self.x_col == self._idxname:
            x_data = self._idx
        else:
            x_data = self.tab[self.x_col].data

        if self.y_col == self._idxname:
            y_data = self._idx
        else:
            y_data = self.tab[self.y_col].data

        x_label = self._get_label('x')
        y_label = self._get_label('y')
        dataobj = self.table_viewer.get_dataobj()
        name = "plot_{}".format(time.time())
        #title = self.trunc(dataobj.get('name', 'NoName'))
        title = name

        if self.tab.masked:
            if self.x_col == self._idxname:
                x_mask = np.ones_like(self._idx, dtype=bool)
            else:
                x_mask = ~self.tab[self.x_col].mask

            if self.y_col == self._idxname:
                y_mask = np.ones_like(self._idx, dtype=bool)
            else:
                y_mask = ~self.tab[self.y_col].mask

            mask = x_mask & y_mask
            x_data = x_data[mask]
            y_data = y_data[mask]

        if len(x_data) > 1:
            i = np.argsort(x_data)  # Sort X-axis to avoid messy line plot
            x_data = x_data[i]
            y_data = y_data[i]

        plot = Plotable(logger=self.logger)
        plot.set_titles(x_axis=x_label, y_axis=y_label, title=title)
        plot.set_grid(True)

        plot.set(name=name, path=None, nothumb=False)

        canvas = plot.get_canvas()
        dc = get_canvas_types()
        points = np.array((x_data, y_data)).T
        canvas.add(dc.Path(points, color='black', linewidth=1,
                           alpha=1.0),
                   redraw=False)

        self.channel.add_image(plot)

    def __str__(self):
        return 'info_ginga_table'



import sys

import numpy as np

import ginga.toolkit as ginga_toolkit
from ginga.misc import log

# window size
dims = (600, 440)


def main(options, args):
    logger = log.get_logger("sine", options=options)

    if options.toolkit is None:
        logger.error("Please choose a GUI toolkit with -t option")

    # decide our toolkit, then import
    ginga_toolkit.use(options.toolkit)

    # now we can import
    from ginga.gw import Widgets, Viewers
    from ginga.canvas.types import plots as gplots
    from ginga.plot.plotaide import PlotAide

    app = Widgets.Application(logger=logger)

    def quit(*args):
        logger.info("Top window closed.")
        sys.exit()

    w = app.make_window("Sine")
    w.add_callback('close', quit)
    app.add_callback('shutdown', quit)

    vbox = Widgets.VBox()
    vbox.set_spacing(1)
    vbox.set_border_width(2)
    win_wd, win_ht = dims[:2]
    viewer = Viewers.CanvasView(logger, render='widget')
    viewer.set_desired_size(win_wd, win_ht)
    viewer.set_zoom_algorithm('rate')
    viewer.set_zoomrate(1.41)
    viewer.enable_autozoom('off')
    viewer.set_background('white')
    viewer.set_foreground('black')
    viewer.set_enter_focus(True)

    viewer.set_limits([(-2, -1), (2, 1)])

    # our plot assistant
    pa = PlotAide(viewer)
    pa.setup_standard_frame(title="Functions", x_title='x', y_title='f(x)')
    pa.settings['autoaxis_x'] = 'off'

    # add scrollbar interface around this viewer
    si = Viewers.GingaScrolledViewerWidget(viewer=viewer, width=win_wd,
                                           height=win_ht)
    pa.configure_scrollbars(si)

    vbox.add_widget(si, stretch=1)

    hbox = Widgets.HBox()
    btn = Widgets.Button("Quit")
    btn.add_callback('activated', quit)
    hbox.add_widget(Widgets.Label(''), stretch=1)
    hbox.add_widget(btn, stretch=0)
    vbox.add_widget(hbox, stretch=0)

    w.set_widget(vbox)

    plot = gplots.CalcPlot(name='Sin',
                           y_fn=np.sin, color='red', linewidth=2.0)
    pa.add_plot(plot)
    plot = gplots.CalcPlot(name='Cos',
                           y_fn=np.cos, color='blue', linewidth=2.0)
    pa.add_plot(plot)

    w.resize(win_wd, win_ht)
    w.show()
    w.raise_()

    pa.zoom_limit_x(-2.0, 2.0)
    pa.update_plots()

    try:
        app.mainloop()

    except KeyboardInterrupt:
        w.close()


if __name__ == "__main__":

    # Parse command line options
    from argparse import ArgumentParser

    argprs = ArgumentParser()

    argprs.add_argument("--debug", dest="debug", default=False,
                        action="store_true",
                        help="Enter the pdb debugger on main()")
    argprs.add_argument("-t", "--toolkit", dest="toolkit", metavar="NAME",
                        default='qt',
                        help="Choose GUI toolkit (gtk|qt)")
    argprs.add_argument("--profile", dest="profile", action="store_true",
                        default=False,
                        help="Run the profiler on main()")
    argprs.add_argument("-r", "--renderer", dest="renderer", metavar="NAME",
                        default=None,
                        help="Choose renderer (pil|agg|opencv|cairo|qt)")
    log.addlogopts(argprs)

    (options, args) = argprs.parse_known_args(sys.argv[1:])

    # Are we debugging this?
    if options.debug:
        import pdb

        pdb.run('main(options, args)')

    # Are we profiling this?
    elif options.profile:
        import profile

        print(("%s profile:" % sys.argv[0]))
        profile.run('main(options, args)')

    else:
        main(options, args)



#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
"""
These stages are all used in implementing the RGB mapping pipeline.
See ginga.RGBMap and ginga.util.stages.render to see how they are used.

Generally speaking, the pipeline looks like

    RGBInput | Distribute | ShiftMap | IntensityMap | ColorMap

"""
import numpy as np

from ginga import ColorDist, trcalc
from ginga import cmap as mod_cmap
from ginga import imap as mod_imap

from .base import Stage, StageError


class RGBMapError(StageError):
    pass


class RGBMapStage(Stage):
    """A class that all stages participating in color mapping subclass.

    Establishes the bit depth of the RGB mapping.
    """
    _stagename = 'rgbmap-base'

    def __init__(self, bpp=8):
        super().__init__()

        self.trace = False
        self.set_bpp(bpp)

    def _set_dtype(self):
        if self.bpp <= 8:
            self.dtype = np.dtype(np.uint8)
        elif self.bpp <= 16:
            self.dtype = np.dtype(np.uint16)
        else:
            self.dtype = np.dtype(np.uint32)

    def set_bpp(self, bpp):
        """
        Set the bit depth (per band) for the output.
        A typical "32-bit RGBA" image would be 8; a 48-bit image would
        be 16, etc.
        """
        self.bpp = bpp
        self.maxc = int(2 ** self.bpp - 1)
        self._set_dtype()


class RGBInput(RGBMapStage):
    """First stage of a sub-pipeline that accepts the input to be colored
    this should usually be the output of the "Cuts" stage of rendering.
    """
    _stagename = 'rgbmap-input'

    def __init__(self, bpp=8):
        super().__init__(bpp=bpp)

        self.in_arr = None

    def set_input(self, in_arr):
        if not np.issubdtype(in_arr.dtype, np.uint):
            in_arr = in_arr.astype(np.uint)

        #self.verify_2d(in_arr)

        self.in_arr = in_arr

    def run(self, prev_stage):
        if prev_stage is not None:
            raise StageError("'{}' in wrong location".format(self._stagename))

        if self._bypass:
            self.pipeline.send(res_np=None)
            return

        if self.trace:
            self.logger.debug("{}: sending data ({})".format(self._stagename,
                                                             self.in_arr))
        self.pipeline.send(res_np=self.in_arr)


class Distribute(RGBMapStage):
    """Distribute data according to a curve.

    This stage handles the distribution of input values according to a
    predefined mapping such as linear, log, power, sqrt, asinh, etc.
    """
    _stagename = 'rgbmap-distribute'

    def __init__(self, bpp=8):
        super().__init__(bpp=bpp)

        self.dist = None
        self._hashsize = self.maxc + 1
        self.set_color_algorithm('linear')

    def set_dist(self, dist):
        self.dist = dist

    def get_dist(self):
        return self.dist

    def get_hash_size(self):
        return self._hashsize

    def set_hash_size(self, size):
        self._hashsize = size
        self.dist.set_hash_size(size)

    def set_color_algorithm(self, name):
        color_dist_class = ColorDist.get_dist(name)
        self.dist = color_dist_class(self._hashsize)

    def get_hasharray(self, arr_in):
        if self._bypass or arr_in is None:
            return arr_in

        if not np.issubdtype(arr_in.dtype, np.uint):
            arr_in = arr_in.astype(np.uint)

        return self.dist.hash_array(arr_in)

    def run(self, prev_stage):
        arr_in = self.pipeline.get_data(prev_stage)
        if self._bypass or arr_in is None:
            self.pipeline.send(res_np=arr_in)
            return
        #self.verify_2d(arr_in)

        arr_out = self.get_hasharray(arr_in)

        if self.trace:
            self.logger.debug("{}: sending result ({})".format(self._stagename,
                                                               arr_out))
        self.pipeline.send(res_np=arr_out)


class ShiftMap(RGBMapStage):
    """Redistribute data according to a shift array.

    This stage handles colormap shifts and stretches, which are used to
    set brightness and contrast in the image.
    """
    _stagename = 'rgbmap-shift-map'

    def __init__(self, bpp=8):
        super().__init__(bpp=bpp)

        self._contrast = 0.5
        self._brightness = 0.5
        self._sarr = None

        self._reset_sarr()

    @property
    def contrast(self):
        return self._contrast

    def set_contrast(self, pct):
        self._contrast = pct
        scale_factor = self._contrast_to_scale(pct)
        self._stretch(scale_factor)

    @property
    def brightness(self):
        return self._brightness

    def set_brightness(self, pct):
        self._brightness = pct
        shift_pct = self._brightness_to_shift(pct)
        self._shift(shift_pct)

    def _scale_to_contrast(self, scale_factor, reset=False):
        if not reset:
            # if not reset, multipy by the current scale factor
            scale_factor *= self._scale_factor
        pct = np.clip((np.clip(scale_factor, 0.01, 20.0) * 1000) / 2000,
                      0.0, 1.0)
        pct = 1.0 - pct
        return pct

    def _contrast_to_scale(self, contrast_pct):
        # I tried to mimic ds9's exponential scale feel along the Y-axis
        i = np.clip(contrast_pct, 0.0001, 1.0)
        scale_factor = (1.0 / (i**3)) * 0.0001 + (1.0 / i) - 1.0
        scale_factor = np.clip(scale_factor, 0.0, self.maxc)
        return scale_factor

    def _brightness_to_shift(self, brightness_pct):
        pct = 1.0 - brightness_pct
        # convert to (-1.0, 1.0) for shift pct
        shift_pct = (pct - 0.5) * 2.0
        return shift_pct

    def _shift_to_brightness(self, shift_pct, reset=False):
        if not reset:
            # if not reset, add the current shift pct
            shift_pct += self._shift_pct
        pct = (np.clip(shift_pct, -1.0, 1.0) + 1) * 0.5
        return 1.0 - pct

    def _reset_sarr(self):
        maxlen = self.maxc + 1
        self._contrast = 0.5
        self._brightness = 0.5
        self._sarr = np.arange(maxlen)

    def _get_sarr(self):
        return self._sarr

    def _set_sarr(self, sarr):
        if sarr is not None:
            sarr = np.asarray(sarr).clip(0, self.maxc).astype(np.uint)
            maxlen = self.maxc + 1
            _len = len(sarr)
            if _len != maxlen:
                raise RGBMapError("shift map length %d != %d" % (_len, maxlen))
            self._sarr = sarr
        else:
            self._reset_sarr()

    def _shift_arr(self, sarr, pct, rotate=False):
        n = len(sarr)
        num = int(n * pct)
        arr = np.roll(sarr, num)
        if not rotate:
            if num > 0:
                arr[0:num] = sarr[0]
            elif num < 0:
                arr[n + num:n] = sarr[-1]
        return arr

    def _stretch_arr(self, sarr, scale):
        old_wd = len(sarr)
        new_wd = int(round(scale * old_wd))

        # Is there a more efficient way to do this?
        xi = np.mgrid[0:new_wd]
        iscale_x = float(old_wd) / float(new_wd)

        xi = (xi * iscale_x).astype(np.uint, copy=False)
        xi = xi.clip(0, old_wd - 1).astype(np.uint, copy=False)
        newdata = sarr[xi]
        return newdata

    def _scale_and_shift(self, scale_factor, shift_pct, callback=True):
        """Stretch and/or shrink the color map via altering the shift map.
        """
        # reset the shift array to normal
        maxlen = self.maxc + 1
        self._sarr = np.arange(maxlen)

        # limit shrinkage to 1% of original size
        scale = max(scale_factor, 0.01)

        work = self._stretch_arr(self._sarr, scale)
        n = len(work)
        if n < maxlen:
            # pad on the lowest and highest values of the shift map
            m = (maxlen - n) // 2 + 1
            barr = np.array([0] * m)
            tarr = np.array([self.maxc] * m)
            work = np.concatenate([barr, work, tarr])
            work = work[:maxlen]

        # we are mimicing ds9's stretch and shift algorithm here.
        # ds9 seems to cut the center out of the stretched array
        # BEFORE shifting
        n = len(work) // 2
        halflen = maxlen // 2
        work = work[n - halflen:n + halflen].astype(np.uint, copy=False)
        assert len(work) == maxlen, \
            RGBMapError("scaled shift map is != %d" % maxlen)

        # shift map according to the shift_pct
        work = self._shift_arr(work, shift_pct)
        assert len(work) == maxlen, \
            RGBMapError("shifted shift map is != %d" % maxlen)

        self._sarr = work

    def _stretch(self, scale_factor, callback=True):
        """Stretch the color map via altering the shift map.
        """
        shift_pct = self._brightness_to_shift(self._brightness)
        self._scale_and_shift(scale_factor, shift_pct, callback=callback)

    def _shift(self, shift_pct, rotate=False, callback=True):
        scale_factor = self._contrast_to_scale(self._contrast)
        self._scale_and_shift(scale_factor, shift_pct, callback=callback)

    def run(self, prev_stage):
        arr_in = self.pipeline.get_data(prev_stage)
        if self._bypass or arr_in is None:
            self.pipeline.send(res_np=arr_in)
            return
        #self.verify_2d(arr_in)

        if not np.issubdtype(arr_in.dtype, np.uint):
            arr_in = arr_in.astype(np.uint)

        # run it through the shift array and clip the result
        # See NOTE [A]
        # arr_out = self._sarr[arr_in].clip(0, self.maxc).astype(np.uint, copy=False)
        arr_out = self._sarr[arr_in]
        # TODO: I think we can avoid this operation, if shift array contents
        # can be limited to 0..maxc
        #arr_out.clip(0, self.maxc, out=arr_out)

        if self.trace:
            self.logger.debug("{}: sending data ({})".format(self._stagename,
                                                             arr_out))
        self.pipeline.send(res_np=arr_out)


class IntensityMap(RGBMapStage):
    """Map indexes through a curve.

    """
    _stagename = 'rgbmap-intensity-map'

    def __init__(self, bpp=8):
        super().__init__(bpp=bpp)

        self.imap = None
        self._iarr = None
        self.set_intensity_map('ramp')

    def set_intensity_map(self, imap_name):
        im = mod_imap.get_imap(imap_name)
        self.set_imap(im)

    def set_imap(self, imap):
        self.imap = imap
        self.calc_imap()

    def calc_imap(self):
        arr = np.array(self.imap.ilst) * float(self.maxc)
        self._iarr = np.round(arr).astype(np.uint, copy=False)

    def run(self, prev_stage):
        arr_in = self.pipeline.get_data(prev_stage)
        if self._bypass or arr_in is None:
            self.pipeline.send(res_np=arr_in)
            return
        #self.verify_2d(arr_in)

        if not np.issubdtype(arr_in.dtype, np.uint):
            arr_in = arr_in.astype(np.uint)

        arr_out = self._iarr[arr_in]

        if self.trace:
            self.logger.debug("{}: sending data ({})".format(self._stagename,
                                                             arr_out))
        self.pipeline.send(res_np=arr_out)


class ColorMap(RGBMapStage):
    """Map indexes through a three-color lookup table.

    """
    _stagename = 'rgbmap-color-map'

    def __init__(self, bpp=8):
        super().__init__(bpp=bpp)

        self.cmap = None
        self._carr = None
        self._inverted = False
        self._rot_pct = 0.0

        self.set_color_map('gray')

    @property
    def inverted(self):
        return self._inverted

    @property
    def rot_pct(self):
        return self._rot_pct

    def set_color_map(self, cmap_name):
        cm = mod_cmap.get_cmap(cmap_name)
        self.set_cmap(cm)

    def set_cmap(self, cmap):
        self.cmap = cmap
        self.calc_cmap()

    def invert_cmap(self, tf):
        self._inverted = tf
        self.calc_cmap()

    def rotate_color_map(self, pct):
        self._rot_pct = np.clip(pct, -1.0, 1.0)
        self.calc_cmap()

    def _gen_cmap(self):
        clst = self.cmap.clst
        self.maxc = len(clst) - 1
        arr = np.array(clst).transpose() * float(self.maxc)
        # does this really need to be the same type as rgbmap output type?
        carr = np.round(arr).astype(self.dtype, copy=False)
        return carr

    def reset_cmap(self):
        self._inverted = False
        self._rot_pct = 0.0
        self.calc_cmap()

    def calc_cmap(self):
        self._carr = self._gen_cmap()
        if self._inverted:
            self._carr = np.fliplr(self._carr)
        num = int((self.maxc + 1) * self._rot_pct)
        if num != 0:
            self._carr = np.roll(self._carr, num, axis=1)

    def get_order_indexes(self, order, cs):
        order = order.upper()
        if order == '':
            # assume standard RGB order if we don't find an order
            # explicitly set
            return [0, 1, 2]
        cs = cs.upper()
        return [order.index(c) for c in cs]

    def do_map_index(self, arr_in, map_arr):
        if not np.issubdtype(arr_in.dtype, np.uint):
            arr_in = arr_in.astype(np.uint)

        # prepare output array
        shape = arr_in.shape
        # get RGB order requested for output
        state = self.pipeline.get('state')
        order = state.order
        depth = len(order)
        image_order = trcalc.guess_order(shape)

        if (image_order is not None) and (len(image_order) > 1):
            # indexes contain RGB axis, so omit this
            res_shape = shape[:-1] + (depth, )
        else:
            res_shape = shape + (depth, )

        out = np.empty(res_shape, dtype=self.dtype, order='C')

        from ginga.RGBMap import RGBPlanes
        rgbobj = RGBPlanes(out, order)

        # set alpha channel
        if rgbobj.hasAlpha:
            aa = rgbobj.get_slice('A')
            aa.fill(self.maxc)

        ri, gi, bi = self.get_order_indexes(rgbobj.get_order(), 'RGB')
        arr_out = rgbobj.rgbarr

        # See NOTE [A]
        if (image_order is None) or (len(image_order) == 1):
            arr_out[..., [ri, gi, bi]] = map_arr[arr_in]

        elif len(image_order) == 2:
            mj, aj = self.get_order_indexes(image_order, 'MA')
            arr_out[..., [ri, gi, bi]] = map_arr[arr_in[..., mj]]

        else:
            # <== indexes already contain RGB info.
            rj, gj, bj = self.get_order_indexes(image_order, 'RGB')
            arr_out[..., ri] = map_arr[:, 0][arr_in[..., rj]]
            arr_out[..., gi] = map_arr[:, 1][arr_in[..., gj]]
            arr_out[..., bi] = map_arr[:, 2][arr_in[..., bj]]

        # alpha = self.pipeline.get('alpha')
        # if alpha is not None:
        #     arr_out[..., -1] = alpha

        return arr_out

    def run(self, prev_stage):
        arr_in = self.pipeline.get_data(prev_stage)
        if self._bypass or arr_in is None:
            self.pipeline.send(res_np=arr_in)
            return
        #self.verify_2d(arr_in)

        arr_out = self.do_map_index(arr_in, self._carr.T)

        if self.trace:
            self.logger.debug("{}: sending data ({})".format(self._stagename,
                                                             arr_out))
        self.pipeline.send(res_np=arr_out)



#
# meta.py -- special mode for launching other modes
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
"""Meta Mode.

Enter the mode by
-----------------
* Space

Exit the mode by
----------------
* Esc

Default bindings in mode
------------------------
* escape : exit any mode
* L : toggle mode lock/oneshot
* S : save image profile

"""
from ginga.modes.mode_base import Mode


class MetaMode(Mode):

    def __init__(self, viewer, settings=None):
        super().__init__(viewer, settings=settings)

        self.actions = dict(
            # Mode 'meta' is special: it is an intermediate mode that
            # is used primarily to launch other modes
            dmod_meta=['space', None, None],

            # KEYBOARD
            kp_save_profile=['S'],
            kp_reset=['escape'],
            kp_lock=['L', 'meta+L'],

            # MOUSE/BUTTON
            ms_none=['nobtn'],
            ms_cursor=['left'],
            ms_wheel=[])

    def __str__(self):
        return 'meta'

    def start(self):
        pass

    def stop(self):
        self.onscreen_message(None)

    def kp_reset(self, viewer, event, data_x, data_y):
        event.accept()
        bm = viewer.get_bindmap()
        bm.reset_mode(viewer)
        viewer.onscreen_message(None)
        return True

    def _toggle_lock(self, viewer, mode_type):
        bm = viewer.get_bindmap()
        # toggle default mode type to locked/oneshot
        dfl_modetype = bm.get_default_mode_type()
        # get current mode
        mode_name, cur_modetype = bm.current_mode()

        if dfl_modetype in ['locked']:
            if mode_type == dfl_modetype:
                mode_type = 'oneshot'

        # install the lock type
        bm.set_default_mode_type(mode_type)
        bm.set_mode(mode_name, mode_type=mode_type)

    def kp_lock(self, viewer, event, data_x, data_y):
        event.accept()
        self._toggle_lock(viewer, 'locked')

    def kp_save_profile(self, viewer, event, data_x, data_y, msg=True):
        event.accept()
        viewer.checkpoint_profile()
        if msg:
            viewer.onscreen_message("Profile saved", delay=0.5)
